from .core import functional
from functools import partial
from jax import numpy as jnp, lax
from typing import Optional, Callable


def spectrogram_helper(
        pad: int,
        n_fft: int,
        hop_length: int,
        win_length: int,
        power: Optional[float] = 2.0,
        normalized: bool = False,
        window_type: Callable = jnp.hanning,
        center: bool = True,
):
    window = window_type(win_length)
    return partial(functional.spectrogram, pad=pad, window=window, n_fft=n_fft,
                   hop_length=hop_length, win_length=win_length, power=power,
                   normalized=normalized, center=center, onesided=True)


def melscale_helper(
        n_mels: int,
        n_fft: int,
        sample_rate: int,
        f_min: float,
        f_max: float,
        norm=None,
        precision=lax.Precision.HIGH
):
    fb = functional.melscale_fbanks(n_freqs=(n_fft//2)+1, n_mels=n_mels,
                         sample_rate=sample_rate, f_min=f_min, f_max=f_max, norm=norm)
    return partial(functional.apply_melscale, melscale_filterbank=fb, precision=precision)



"""
Helper functions for jax/flax training

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import os
import json
import logging
import functools
import ml_collections
import optax
from typing import Any
import jax
from jax import lax, random
import flax
from flax import jax_utils
from jax import numpy as jnp
from flax.training import checkpoints
from flax.training import train_state
from flax.training import common_utils
from flax import optim
from .misc import TrainingMode, Features, DataSplit
from flax.core import freeze, unfreeze
from .. import feature_helper
from .trainstate import TrainState_v2
from sklearn.metrics import average_precision_score
from .. import models
from .. import frontends
from ..transforms.spec_augment import SpecAugment


def get_dtype(half_precision: bool):
    platform = jax.local_devices()[0].platform
    if half_precision:
        if platform == 'tpu':
            model_dtype = jnp.bfloat16
        else:
            model_dtype = jnp.float16
    else:
        model_dtype = jnp.float32
    return model_dtype


def restore_checkpoint(state, workdir):
    return checkpoints.restore_checkpoint(workdir, state)


def save_checkpoint(state, workdir):
    if jax.process_index() == 0:
        # get train state from the first replica
        state = jax.device_get(jax.tree_map(lambda x: x[0], state))
        step = int(state.step)
        checkpoints.save_checkpoint(workdir, state, step, keep=3)


def save_best_checkpoint(state, workdir, best_acc):
    if jax.process_index() == 0:
        # get train state from the first replica
        state = jax.device_get(jax.tree_map(lambda x: x[0], state))
        step = int(state.step)
        checkpoints.save_checkpoint(workdir, state, step, prefix="best_ckpt_", keep=3)


def initialize(key, inp_shape, model, 
               aux_rng_keys=["dropout", "drop_path", "mixup", "spec_aug"]):
    input_shape = (2,) + inp_shape

    @jax.jit
    def init(*args):
        return model.init(*args)
    num_keys = len(aux_rng_keys)
    key, *subkeys = jax.random.split(key, num_keys+1)
    rng_keys = {aux_rng_keys[ix]: subkeys[ix] for ix in range(len(aux_rng_keys))}
    variables = init({'params': key, **rng_keys}, 
                     jnp.ones(input_shape, model.dtype))
    rngs = flax.core.FrozenDict(rng_keys)
    if "batch_stats" in variables.keys():
        return variables['params'], variables['batch_stats'], rngs
    else:
        return variables['params'], flax.core.freeze({}), rngs


# pmean only works inside pmap because it needs an axis name.
# This function will average the inputs across all devices.
cross_replica_mean = jax.pmap(lambda x: lax.pmean(x, 'x'), 'x')


def sync_batch_stats(state):
    """Sync the batch statistics across replicas."""
    # Each device has its own version of the running average batch statistics and
    # we sync them before evaluation.
    if len(state.batch_stats) != 0:
        return state.replace(batch_stats=cross_replica_mean(state.batch_stats))
    else:
        return state


def create_learning_rate_fn(
        config: ml_collections.ConfigDict,
        base_learning_rate: float,
        steps_per_epoch: int):
    """Create learning rate schedule."""
    warmup_fn = optax.linear_schedule(
        init_value=0., end_value=base_learning_rate,
        transition_steps=config.opt.warmup_epochs * steps_per_epoch)
    cosine_epochs = max(config.num_epochs - config.opt.warmup_epochs, 1)
    cosine_fn = optax.cosine_decay_schedule(
        init_value=base_learning_rate,
        decay_steps=cosine_epochs * steps_per_epoch)
    schedule_fn = optax.join_schedules(
        schedules=[warmup_fn, cosine_fn],
        boundaries=[config.opt.warmup_epochs * steps_per_epoch])
    return schedule_fn


def prepare_tf_data(xs, devices=None):
    """Convert a input batch from tf Tensors to numpy arrays."""
    if devices is None:
        local_device_count = jax.local_device_count()
    elif type(devices) == list or type(devices) == tuple:
        local_device_count = len(devices)
    else:
        raise ValueError("Devices should either be None or a list of jax.device")

    def _prepare(x):
        # Use _numpy() for zero-copy conversion between TF and NumPy.
        x = x._numpy()  # pylint: disable=protected-access

        # reshape (host_batch_size, height, width, channel) to
        # (local_devices, device_batch_size, height, width, channel)
        return x.reshape((local_device_count, -1) + x.shape[1:])

    return jax.tree_map(_prepare, xs)


def create_train_state(rng, config: ml_collections.ConfigDict,
                       model, learning_rate_fn):
    """Create initial training state."""
    dynamic_scale = None
    platform = jax.local_devices()[0].platform
    if config.half_precision and platform == 'gpu':
        dynamic_scale = optim.DynamicScale()
    else:
        dynamic_scale = None

    params, batch_stats, rng_keys = initialize(rng, config.input_shape, model)
    tx = optax.adamw(
        learning_rate=learning_rate_fn,
        weight_decay=config.opt.weight_decay
    )
    state = TrainState_v2.create(
        apply_fn=model.apply,
        params=params,
        frozen_params=flax.core.freeze({}),
        tx=tx,
        batch_stats=batch_stats,
        aux_rng_keys=rng_keys,
        dynamic_scale=dynamic_scale)
    return state


def create_train_state_from_pretrained(rng, config: ml_collections.ConfigDict,
                                       model, learning_rate_fn,
                                       pretrained_work_dir, pretrained_prefix="checkpoint_",
                                       to_copy=['encoder'],
                                       fc_only=False):
    logging.info("Making train state from pretrained..")
    dynamic_scale = None
    platform = jax.local_devices()[0].platform
    if config.half_precision and platform == 'gpu':
        dynamic_scale = optim.DynamicScale()
    else:
        dynamic_scale = None
    params, batch_stats, rng_keys = initialize(rng, config.input_shape, model)

    # load pretrained ckpt to a dictionary
    pretrained_state_dict = checkpoints.restore_checkpoint(pretrained_work_dir, None,
                                                           prefix=pretrained_prefix)
    pretrained_params = pretrained_state_dict['params']
    pretrained_batch_stats = pretrained_state_dict['batch_stats']

    # unfreeze classifier params and batch_stats
    params = unfreeze(params)
    batch_stats = unfreeze(batch_stats)

    # copy stuff
    for k in to_copy:
        assert k in params.keys() #and k in batch_stats.keys()
        params[k] = pretrained_params[k]
        try:
            batch_stats[k] = pretrained_batch_stats[k]
        except KeyError as ex:
            pass

    # filter params based on fc_only argument
    if fc_only:
        logging.info("Finetuning fc-only layer")
        frozen_params = {}
        trainable_params = {}
        for k in params.keys():
            if k in to_copy:
                frozen_params[k] = params[k]
            else:
                trainable_params[k] = params[k]
    else:
        frozen_params = {}
        trainable_params = params

    # freeze classifier params and batch_stats
    trainable_params = freeze(trainable_params)
    frozen_params = freeze(frozen_params)
    batch_stats = freeze(batch_stats)

    # make the train state now
    tx = optax.adamw(
        learning_rate=learning_rate_fn,
        weight_decay=config.opt.weight_decay
    )
    state = TrainState_v2.create(
        apply_fn=model.apply,
        params=trainable_params,
        frozen_params=frozen_params,
        tx=tx,
        batch_stats=batch_stats,
        aux_rng_keys=rng_keys,
        dynamic_scale=dynamic_scale)
    return state


def cross_entropy_loss(logits, labels, smoothing_factor: float = None):
    if smoothing_factor and type(smoothing_factor) == float:
        labels = optax.smooth_labels(labels, alpha=smoothing_factor)
    xentropy = optax.softmax_cross_entropy(logits=logits, labels=labels)
    return jnp.mean(xentropy)


def binary_xentropy_loss(logits, labels, smoothing_factor: float = None):
    if smoothing_factor and type(smoothing_factor) == float:
        labels = optax.smooth_labels(labels, alpha=smoothing_factor)
    xentropy = optax.sigmoid_binary_cross_entropy(logits=logits, labels=labels)
    return jnp.mean(xentropy)


def compute_metrics(logits, labels, mode, cost_fn):
    loss = cost_fn(logits, labels)
    metrics = {
        'loss': loss
    }
    if mode == TrainingMode.MULTICLASS:
        accuracy = jnp.mean(jnp.argmax(logits, -1) == jnp.argmax(labels, -1))
        # labels is now a (batch, num_classes) onehot-encoded array
        metrics['accuracy'] = accuracy

    metrics = lax.pmean(metrics, axis_name='batch')
    return metrics


def create_input_iter(ds, devices=None):
    ds = ds.repeat()
    prep_data = functools.partial(prepare_tf_data, devices=devices)
    it = map(prep_data, ds)
    it = jax_utils.prefetch_to_device(it, 10, devices=devices)
    return it


def apply_audio_transforms(batch, transforms=[],
                           dtype=jnp.float32, normalize=False):
    output = batch
    for tfs in transforms:
        output = tfs(output)
    output = jnp.clip(output, a_min=1e-8, a_max=1e8)
    if len(transforms) != 0:
        output = jnp.log(output)
    if normalize:
        print("!!!!!!!!!!!!!!!!!!!!!! NORMALIZING INPUTS !!!!!!!!!!!!!!!!!!!!!!!!")
        mu = output.mean(axis=(1, 2), keepdims=True)
        std = output.std(axis=(1, 2), keepdims=True)
        output = (output - mu)/(std+1e-7)
    output = output[Ellipsis, jnp.newaxis]
    # output = output.transpose((0, 2, 1, 3))
    output = output.astype(dtype)
    return output


def get_feature_functions(config: ml_collections.ConfigDict):
    if not config.data.jax_transforms:
        return None
    
    if config.audio_config.features == "raw":
        tfs = []
    else:
        tfs = []
        tfs.append(feature_helper.spectrogram_helper(
            pad=0,
            n_fft=config.audio_config.n_fft,
            hop_length=config.audio_config.hop_len,
            win_length=config.audio_config.win_len
        ))
        if config.audio_config.features == "log_mel":
            tfs.append(feature_helper.melscale_helper(
                n_mels=config.audio_config.n_mels,
                n_fft=config.audio_config.n_fft,
                sample_rate=config.audio_config.sample_rate,
                f_min=config.audio_config.fmin,
                f_max=config.audio_config.fmax,
                precision=lax.Precision.HIGHEST
            ))
    return tfs


def get_eval_metrics_summary(eval_metrics, eval_logits, eval_labels, mode=TrainingMode.MULTICLASS):
    eval_metrics = common_utils.get_metrics(eval_metrics)

    summary = jax.tree_map(lambda x: x.mean(), eval_metrics)
    if mode == TrainingMode.MULTICLASS:
        return summary
    else:
        # if mode == TrainingMode.MULTICLASS:
        #     logging.info('eval epoch: %d, loss: %.4f, accuracy: %.4f',
        #                  epoch, summary['loss'], summary['accuracy'] * 100)
        eval_logits = jnp.concatenate([jax.device_get(x) for x in eval_logits])
        eval_labels = jnp.concatenate([jax.device_get(x) for x in eval_labels])

        eval_logits = eval_logits.reshape(-1, eval_logits.shape[-1])
        eval_labels = eval_labels.reshape(-1, eval_labels.shape[-1])
        # print("logits.shape: {} | type: {}".format(eval_logits.shape, type(eval_logits)))

        map_value = average_precision_score(eval_labels.astype('float32'), eval_logits.astype('float32'),
                                            average="macro")

        summary['accuracy'] = map_value
        return summary
        # logging.info('eval epoch: %d, loss: %.4f, accuracy: %.4f',
        #              epoch, summary['loss'], summary['accuracy'] * 100)


def read_config_from_json(workdir):
    config_path = os.path.join(workdir, "config.json")
    print(f"loading config -> {config_path}")
    with open(config_path, "r") as fd:
        config_dict = json.load(fd)
        config_dict['input_shape'] = tuple(config_dict['input_shape'])
    config = ml_collections.ConfigDict(config_dict)
    return config


def write_config_to_json(workdir, config):
    config_path = os.path.join(workdir, "config.json")
    if not os.path.exists(workdir):
        os.makedirs(workdir)
    if os.path.exists(config_path):
        logging.info(f"config file {config_path} exists.. Not overwriting.")
        return
    with open(config_path, "w") as fd:
        json.dump(config.to_dict(), fd)


def get_model_frontend_cls(config):
    model_cls = getattr(models, config.model.arch)
    model_args = config.model.get("model_args", None)
    if model_args:
        model_cls = functools.partial(model_cls, **model_args.to_dict())
    # if "efficient" in config.model.arch:
    #     model_cls = functools.partial(model_cls, drop_path_prob=0.)
    frontend = config.model.get("frontend", None)
    frontend_args = config.model.get("frontend_args", None)
    if frontend:
        frontend_cls = getattr(frontends, frontend)
        if frontend_args:
            frontend_cls = functools.partial(frontend_cls, **frontend_args.to_dict())
    else:
        frontend_cls = None
    return model_cls, frontend_cls


def get_spec_augment(config, devices=None, pmapped=False):
    if config.opt.get("spec_augment", False):
        spec_aug_args = config.opt.get("spec_augment_args", None)
        if spec_aug_args:
            spec_aug_func = SpecAugment(**spec_aug_args.to_dict())
        else:
            spec_aug_func = SpecAugment(20, 50)

        if pmapped:
            p_spec_aug = jax.pmap(
                spec_aug_func, axis_name="batch", devices=devices
            )
        else:
            p_spec_aug = spec_aug_func
        spec_aug_rng = random.PRNGKey(9991).reshape(-1, 2)
        spec_aug_rng = jnp.repeat(spec_aug_rng, len(devices), axis=0)
    else:
        p_spec_aug = None
        spec_aug_rng = None
    return  p_spec_aug, spec_aug_rng



# credits: https://github.com/hche11
# https://github.com/hche11/VGGSound/blob/master/utils.py
# Modifications and additions for audax by / Copyright 2022, Sarthak Yadav

import csv
import torch
from sklearn.metrics import average_precision_score
from scipy import stats
from sklearn import metrics
import numpy as np


def accuracy(output, target, topk=(1, 5)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res, pred


def d_prime(auc):
    standard_normal = stats.norm()
    d_prime = standard_normal.ppf(auc) * np.sqrt(2.0)
    return d_prime


def calculate_stats(output, target, class_indices=None):
    """Calculate statistics including mAP, AUC, etc.

    Args:
      output: 2d array, (samples_num, classes_num)
      target: 2d array, (samples_num, classes_num)
      class_indices: list
        explicit indices of classes to calculate statistics for

    Returns:
      stats: list of statistic of each class.
    """

    classes_num = target.shape[-1]
    if class_indices is None:
        class_indices = range(classes_num)
    stats = []

    # Class-wise statistics
    for k in class_indices:
        # Average precision
        avg_precision = metrics.average_precision_score(
            target[:, k], output[:, k], average=None)

        # AUC
        auc = metrics.roc_auc_score(target[:, k], output[:, k], average=None)

        # Precisions, recalls
        (precisions, recalls, thresholds) = metrics.precision_recall_curve(
            target[:, k], output[:, k])

        # FPR, TPR
        (fpr, tpr, thresholds) = metrics.roc_curve(target[:, k], output[:, k])

        save_every_steps = 1000     # Sample statistics to reduce size
        dict = {'precisions': precisions[0::save_every_steps],
                'recalls': recalls[0::save_every_steps],
                'AP': avg_precision,
                'fpr': fpr[0::save_every_steps],
                'fnr': 1. - tpr[0::save_every_steps],
                'auc': auc}
        stats.append(dict)

    return stats


def mean_confidence_interval(data, confidence=0.95):
    a = 1.0 * np.array(data)
    n = len(a)
    m, se = np.mean(a), stats.sem(a)
    h = se * stats.t.ppf((1 + confidence) / 2., n-1)
    return m, h


def calculate_mAP(preds, gts, mixup=False, mode="macro"):
    preds = torch.cat(preds, 0).numpy()
    gts = torch.cat(gts, 0).numpy()
    if mixup:
        gts[gts >= 0.5] = 1
        gts[gts < 0.5] = 0
    map_value = average_precision_score(gts, preds, average=mode)
    return map_value


"""
Implementation for the paper

Zhang, H., Cisse, M., Dauphin, Y.N. and Lopez-Paz, D., 2017.
"mixup: Beyond empirical risk minimization". arXiv preprint arXiv:1710.09412.

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import jax
import jax.numpy as jnp
from ..training_utils.misc import TrainingMode


def do_mixup(key, inputs, targets, alpha=0.2, mode: TrainingMode=TrainingMode.MULTILABEL):
    """
    Applies Mixup to the input.
    :param inputs: input batch tensor of shape (bsize, C, H, W)
    :param targets: input gt tensor of shape (bsize, num_classes)
    :param alpha: alpha parameter for beta distribution generation
    :return: tuple of mixed_inputs, mixed_outputs
    """
    key, subkey = jax.random.split(key, 2)
    ndim = inputs.ndim
    bsize = inputs.shape[0]
    lam = jax.random.beta(key, alpha, alpha, shape=(bsize,))
    perms = jax.random.permutation(subkey, bsize)
    
    lam_desired_shape = (bsize,) + (1,) * (ndim-1)
    mixed_x = inputs * lam.reshape(lam_desired_shape) + inputs[perms] * (1 - lam.reshape(lam_desired_shape))

    if mode == TrainingMode.MULTILABEL:
        mixed_y = targets * lam.reshape((bsize, 1)) + targets[perms] * (1 - lam.reshape((bsize, 1)))
        return mixed_x, mixed_y, None, None
    else:
        y_a, y_b = targets, targets[perms]
        return mixed_x, y_a, y_b, lam


def get_mixedup_xentropy(criterion, pred, y_a, y_b, lam, mode: TrainingMode=TrainingMode.MULTILABEL):
    if mode == TrainingMode.MULTILABEL:
        return criterion(pred, y_a)
    elif mode == TrainingMode.MULTICLASS:
        return criterion(pred, y_a) * lam + criterion(pred, y_b) * (1 - lam)
    else:
        raise ValueError(f"Unsupported training mode {mode}, should be one of ['TrainingMode.MULTILABEL', 'TrainingMode.MULTICLASS']")



"""
Implementation of the MLP layer as used in recent literature for vision transformers,etc

Inspired by the PyTorch implementation in timm (https://github.com/rwightman/pytorch-image-models)
by Ross Wightman, 2022

Modified for audax by / Copyright 2022, Sarthak Yadav
"""
import jax.numpy as jnp
import flax.linen as nn
from typing import Any, Callable


default_init = nn.initializers.lecun_normal


class Mlp(nn.Module):
    """
    Mlp as used in Vision transformers, MLP mixers and such
    """
    hidden_features: int
    out_features: int = None
    drop: float = 0.
    activation: Callable = nn.gelu
    dtype: Any = jnp.float32
    kernel_init: Callable = default_init

    @nn.compact
    def __call__(self, inputs, train: bool = True):
        out_features = self.out_features or self.hidden_features
        hidden_feature = self.hidden_features

        outputs = inputs.astype(self.dtype)
        outputs = nn.Dense(hidden_feature, kernel_init=self.kernel_init)(outputs)
        outputs = self.activation(outputs)
        outputs = nn.Dropout(self.drop)(outputs, deterministic=train)
        outputs = nn.Dense(out_features, kernel_init=self.kernel_init)(outputs)
        outputs = nn.Dropout(self.drop)(outputs, deterministic=train)
        return outputs



from .resnet import ResNet18, ResNet34, ResNet50, ResNet101, ResNet152, ResNet200
from .convnext import convnext_tiny, convnext_small, convnext_base, convnext_large, convnext_xlarge
from .efficientnet import efficientnet_b0, efficientnet_b1, efficientnet_b2, efficientnet_b3, efficientnet_b4, efficientnet_b5, efficientnet_b6, efficientnet_b7
from .classifier import Classifier


"""
Similarity layers

Written/modified for audax by / Copyright 2022, Sarthak Yadav
"""
from jax import numpy as jnp, lax
from flax import linen as nn
from typing import Callable, Any, Iterable, Optional
from ..utils import l2_normalize

PRNGKey = Any
Shape = Iterable[int]
Dtype = Any  # this could be a real type?
Array = Any


class DotProduct(nn.Module):
    precision: lax.Precision = lax.Precision.HIGH

    @nn.compact
    def __call__(self, anchor, positive):
        anchor = l2_normalize(anchor, axis=-1)
        positive = l2_normalize(positive, axis=-1)
        return jnp.matmul(anchor, positive.T, precision=self.precision)


class BilinearProduct(nn.Module):
    dim: int
    dtype: Dtype = jnp.float32

    @nn.compact
    def __call__(self, anchor, positive):
        w = self.param('bilinear_product_weight', nn.initializers.normal(), [self.dim, self.dim], self.dtype)
        projection_positive = jnp.matmul(w, positive.T)
        return jnp.matmul(anchor, projection_positive)



"""
Audio dataset and parsing utilities for audax

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import os
import tensorflow as tf
import tensorflow_io as tfio
from typing import Callable, Optional
from functools import partial
from ..misc import TrainingMode, DataSplit
from .transforms import pad_waveform, contrastive_labels


def parse_tfrecord_fn_v2(example,
                         label_parser=None,
                         cropper=None,
                         seg_length=16000):
    feature_description = {
        "audio": tf.io.FixedLenFeature([], tf.string),
        "label": tf.io.VarLenFeature(tf.int64),
        "duration": tf.io.FixedLenFeature([], tf.int64),
    }
    example = tf.io.parse_single_example(example, feature_description)
    a = tf.io.parse_tensor(example['audio'], tf.float32)
    a.set_shape([None])
    a = pad_waveform(a, seg_length=seg_length)
    if cropper:
        a = tf.numpy_function(cropper, [a], tf.float32)
    example['audio'] = a
    if label_parser:
        example = label_parser(example)
    return example


def parse_tfrecord_fn_tfio(example,
                           label_parser=None,
                           cropper=None,
                           normalize_waveform=False,
                           seg_length=16000):
    feature_description = {
        "audio": tf.io.FixedLenFeature([], tf.string),
        "label": tf.io.VarLenFeature(tf.int64),
        "duration": tf.io.FixedLenFeature([], tf.int64),
    }
    example = tf.io.parse_single_example(example, feature_description)
    a = tfio.audio.decode_flac(example['audio'], dtype=tf.int16)
    a = tf.reshape(a, (-1,))
    a = tf.cast(a, tf.float32) / 32768.
    a = pad_waveform(a, seg_length=seg_length)
    if cropper:
        a = cropper(a)
    if normalize_waveform:
        a = tf.math.l2_normalize(a, epsilon=1e-9)
    example['audio'] = a
    if label_parser:
        example = label_parser(example)
    return example


def parse_tf_record_tfio_contrastive(example, cropper,
                                     label_parser=None,
                                     normalize_waveform=True,
                                     noise=0.001,
                                     seg_length=16000):
    feature_description = {
        "audio": tf.io.FixedLenFeature([], tf.string),
        "label": tf.io.VarLenFeature(tf.int64),
        "duration": tf.io.FixedLenFeature([], tf.int64),
    }
    example = tf.io.parse_single_example(example, feature_description)
    a = tfio.audio.decode_flac(example['audio'], dtype=tf.int16)
    a = tf.reshape(a, (-1,))
    a = tf.cast(a, tf.float32) / 32768.
    a = pad_waveform(a, seg_length=seg_length)

    waveform_a = cropper(a)
    waveform_p = cropper(a)

    if normalize_waveform:
        waveform_a = tf.math.l2_normalize(waveform_a, epsilon=1e-9)
        waveform_p = tf.math.l2_normalize(waveform_p, epsilon=1e-9)

    waveform_p = waveform_p + (noise * tf.random.normal(tf.shape(waveform_p)))
    output = {
        "anchor": waveform_a,
        "positive": waveform_p
    }
    return output


def map_contrastive_labels(example):
    labels = tf.range(0, limit=example['anchor'].shape[0], dtype=tf.int64)
    labels = tf.one_hot(labels, depth=example['anchor'].shape[0])
    example['label'] = labels
    return example


def get_dataset_v2(filenames,
                batch_size,
                parse_example,
                num_classes,
                data_split=DataSplit.TRAIN,
                file_ext="tfrec",
                compression="ZLIB",
                feature_extraction_func=None,
                cacheable=False,
                is_contrastive=False
                ):
    options = tf.data.Options()
    options.autotune.enabled = True
    options.threading.private_threadpool_size = 96  # 0=automatically determined
    options.deterministic = False
    dataset = tf.data.Dataset.from_tensor_slices(filenames)
    dataset = dataset.with_options(options)
    # shuffle filenames every epoch
    dataset = dataset.shuffle(len(filenames), seed=0, reshuffle_each_iteration=True)    
    # dataset = tf.data.TFRecordDataset(filenames, compression_type=compression,
    #                                       num_parallel_reads=tf.data.AUTOTUNE)
    # dataset = dataset.with_options(options)
    # dataset = dataset.map(parse_example, num_parallel_calls=tf.data.AUTOTUNE).prefetch(tf.data.AUTOTUNE)
    if cacheable:
        dataset = dataset.interleave(
            lambda x: tf.data.TFRecordDataset(x, compression_type=compression,
                                            num_parallel_reads=tf.data.AUTOTUNE).prefetch(tf.data.AUTOTUNE),
            cycle_length=tf.data.AUTOTUNE, block_length=32,
            num_parallel_calls=tf.data.AUTOTUNE, deterministic=False).prefetch(tf.data.AUTOTUNE)
        dataset = dataset.cache()
        dataset = dataset.map(parse_example, num_parallel_calls=tf.data.AUTOTUNE)
    else:
        # gives better data throughput for AudioSet, which can't really be cached anyway
        dataset = dataset.interleave(
            lambda x: tf.data.TFRecordDataset(x, compression_type=compression,
                                              num_parallel_reads=tf.data.AUTOTUNE).map(parse_example, num_parallel_calls=tf.data.AUTOTUNE),
            cycle_length=tf.data.AUTOTUNE, block_length=32,
            num_parallel_calls=tf.data.AUTOTUNE, deterministic=False).prefetch(tf.data.AUTOTUNE)

    dataset = dataset.batch(batch_size, num_parallel_calls=tf.data.AUTOTUNE, drop_remainder=True)
    if feature_extraction_func:
        dataset = dataset.map(feature_extraction_func,
                              num_parallel_calls=tf.data.AUTOTUNE)
    if is_contrastive:
        dataset = dataset.map(contrastive_labels, num_parallel_calls=tf.data.AUTOTUNE)
    dataset = dataset.prefetch(tf.data.AUTOTUNE)
    return dataset



"""
Tensorflow FFT is painfully slow
https://github.com/tensorflow/tensorflow/issues/6541

This file provides a sample of how to use torch transforms in tf.dataset pipelines. It's much faster than using
tensorflow fft, but much slower than doing everything in jax after batches have been created.

Additional benefit: switching frameworks won't affect input features


The transforms here should be used following these general guidelines


- SpectrogramParser, SpectrogramPostProcess and ToMelScale should be called on batched signals
  of shape (N, T), and thus should be used after batching has been done in your pipeline

- Cropping functions should be called early in the pipeline, on unbatched signals of shape (T,)

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import functools
import torch
from audio_utils.common.transforms import Compose, CenterCrop, RandomCrop
from audio_utils.common.feature_transforms import SpectrogramParser, SpectrogramPostProcess, ToMelScale


def apply_feature_extractor(audios, feature_extractor):
    """
    Applies given feature extractor to a batch of audios

    This function can be easily wrapped in tf.Dataset pipeline as

    >>> extractor_features = functools.partial(apply_feature_extractor, feature_extractor)
    >>> # where `feature_extractor` is the callable transform
    >>> tf.numpy_function(extractor_features, [input_batch], Tout=tf.float32)

    Parameters
    ----------
    audios: float32 array of shape (N, T)
    feature_extractor: callable object for features

    Returns
    -------
    np array of shape (N, frequency_bins, timesteps, 1)
    """
    x = torch.from_numpy(audios).unsqueeze(1)
    return feature_extractor(x).squeeze(1).unsqueeze(-1).numpy()


def apply_cropper(audio, cropper):
    """
    Applies given cropper to a batch of audios

    This function can be easily wrapped in tf.Dataset pipeline as

    >>> crop_features = functools.partial(apply_cropper, cropper)
    >>> # where `cropper` is the callable crop transform
    >>> tf.numpy_function(crop_features, [input_audio], Tout=tf.float32)

    Parameters
    ----------
    audio: float32 waveform array of shape (T,)
    cropper: callable cropper object that returns a (M,) shaped crop of input, M <= T

    Returns
    -------
    float32 np.array of shape (N, M)
    """
    x = torch.from_numpy(audio).unsqueeze(0)
    return cropper(x).squeeze(0).numpy()


def get_spectrogram_extractor(frame_length=400,
                              frame_step=160,
                              fft_length=1024,
                              log_compress=True,
                              normalize=False,
                              power=2,
                              **kwargs):
    tfs = Compose([
        SpectrogramParser(window_length=frame_length, hop_length=frame_step, n_fft=fft_length, mode="after_batch"),
        SpectrogramPostProcess(window_length=frame_length, normalize=normalize,
                               log_compress=log_compress, power=power, mode="after_batch")
    ])
    return functools.partial(apply_feature_extractor, feature_extractor=tfs)


def get_log_mel_extractor(
        sample_rate=16000,
        frame_length=400,
        frame_step=160,
        fft_length=1024,
        n_mels=64,
        fmin=60.0,
        fmax=7800.0,
        power=2,
        normalize=False,
        **kwargs):
    tfs = Compose([
        SpectrogramParser(window_length=frame_length, hop_length=frame_step, n_fft=fft_length, mode="after_batch"),
        SpectrogramPostProcess(window_length=frame_length, power=power, normalize=normalize,
                               log_compress=False, mode="after_batch"),
        ToMelScale(sample_rate=sample_rate, hop_length=frame_step, n_fft=fft_length, n_mels=n_mels,
                   fmin=fmin, fmax=fmax, norm=None)
    ])
    return functools.partial(apply_feature_extractor, feature_extractor=tfs)


def get_random_cropper(seq_length):
    cropper = RandomCrop(seq_length)
    return functools.partial(apply_cropper, cropper=cropper)


def get_center_cropper(seq_length):
    cropper = CenterCrop(seq_length)
    return functools.partial(apply_cropper, cropper=cropper)



"""
Contrastive training utilities

Should work out of the box with any contrastive ssl losses that have anchor-positive pairs.

However, primarily designed to work with SimCLR [1, 2] & COLA [3], so some modifications might be necessary
for other use cases

References
----------

[1] Chen, T., Kornblith, S., Norouzi, M. and Hinton, G., 2020, November.
    "A simple framework for contrastive learning of visual representations".
    In International conference on machine learning (pp. 1597-1607). PMLR.


[2] Chen, T., Kornblith, S., Swersky, K., Norouzi, M. and Hinton, G., 2020.
    "Big self-supervised models are strong semi-supervised learners".
    arXiv preprint arXiv:2006.10029.

[3] Saeed, A., Grangier, D. and Zeghidour, N., 2021, June.
    "Contrastive learning of general-purpose audio representations".
    In ICASSP 2021-2021 IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP)
    (pp. 3875-3879). IEEE.

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import functools
import time
from typing import Any

from absl import logging
from clu import metric_writers
from clu import periodic_actions
import flax
from flax import jax_utils
from flax import optim
from flax.training import checkpoints
from flax.training import common_utils
from flax.training import train_state
import jax
from jax import lax
import jax.numpy as jnp
from jax import random
import ml_collections
import optax
import tensorflow as tf
from sklearn.metrics import average_precision_score

try:
    # Disable all GPUS
    tf.config.set_visible_devices([], 'GPU')
    visible_devices = tf.config.get_visible_devices()
    for device in visible_devices:
        assert device.device_type != 'GPU'
except:
    # Invalid device or cannot modify virtual devices once initialized.
    pass

from .misc import TrainingMode, Features, DataSplit
from . import training_utilities
from .data_v2.helpers import prepare_datasets_v2

# import encoders
# from cola import COLA, SimilarityMeasure
import wandb


# def create_COLA_model(*, encoder_cls, half_precision, similarity_measure, embedding_dim,
#                       temperature=0.1):
#     platform = jax.local_devices()[0].platform
#     if half_precision:
#         if platform == 'tpu':
#             model_dtype = jnp.bfloat16
#         else:
#             model_dtype = jnp.float16
#     else:
#         model_dtype = jnp.float32
#     return COLA(model_cls=encoder_cls, similarity_measure=similarity_measure,
#                 embedding_dim=embedding_dim, temperature=temperature, dtype=model_dtype)


def train_step(state, batch, learning_rate_fn,
               cost_func,
               mode=TrainingMode.MULTICLASS):
    """Perform a single training step."""

    def loss_fn(params):
        anchors = batch['anchor']
        positives = batch['positive']
        labels = batch['label']
        logits, new_model_state = state.apply_fn(
            {'params': params, 'batch_stats': state.batch_stats},
            jnp.concatenate([anchors, positives]),
            mutable=['batch_stats'],
            rngs=state.aux_rng_keys
        )
        labels = common_utils.onehot(jnp.arange(0, logits.shape[0]), num_classes=logits.shape[0])
        loss = cost_func(logits, labels)
        return loss, (new_model_state, logits, labels)

    step = state.step
    dynamic_scale = state.dynamic_scale
    lr = learning_rate_fn(step)

    if dynamic_scale:
        grad_fn = dynamic_scale.value_and_grad(
            loss_fn, has_aux=True, axis_name='batch')
        dynamic_scale, is_fin, aux, grads = grad_fn(state.params)
        # dynamic loss takes care of averaging gradients across replicas
    else:
        grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
        aux, grads = grad_fn(state.params)
        # Re-use same axis_name as in the call to `pmap(...train_step...)` below.
        grads = lax.pmean(grads, axis_name='batch')

    new_model_state, logits, labels = aux[1]
    metrics = training_utilities.compute_metrics(logits, labels, mode=mode, cost_fn=cost_func)
    metrics['learning_rate'] = lr
    new_state = state.apply_gradients(
        grads=grads, batch_stats=new_model_state['batch_stats'])
    if dynamic_scale:
        new_state = new_state.replace(
            opt_state=jax.tree_multimap(
                functools.partial(jnp.where, is_fin),
                new_state.opt_state,
                state.opt_state),
            params=jax.tree_multimap(
                functools.partial(jnp.where, is_fin),
                new_state.params,
                state.params))
        metrics['scale'] = dynamic_scale.scale
    return new_state, metrics


def eval_step(state, batch, cost_func, mode=TrainingMode.MULTICLASS):
    variables = {'params': state.params, 'batch_stats': state.batch_stats}
    anchors = batch['anchor']
    positives = batch['positive']
    logits = state.apply_fn(
        variables, jnp.concatenate([anchors, positives]), train=False, mutable=False)
    labels = batch['label']
    labels = common_utils.onehot(jnp.arange(0, logits.shape[0]), num_classes=logits.shape[0])
    metrics = training_utilities.compute_metrics(logits, labels, mode=mode, cost_fn=cost_func)
    return metrics, logits, labels



"""
Data transforms in tensorflow for the tf.data based datasets
Kept separate from core audax (since it's tensorflow based)

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import tensorflow as tf


def pad_waveform(waveform, seg_length=16000):
    padding = tf.maximum(seg_length - tf.shape(waveform)[0], 0)
    left_pad = padding // 2
    right_pad = padding - left_pad
    padded_waveform = tf.pad(waveform, paddings=[[left_pad, right_pad]])
    return padded_waveform


def random_crop_signal(audio, slice_length):
    data_length = tf.shape(audio, out_type=tf.dtypes.int64)[0]
    max_offset = data_length - slice_length
    if max_offset == 0:
        return audio
    random_offset = tf.random.uniform((), minval=0, maxval=max_offset, dtype=tf.dtypes.int64)
    slice_indices = tf.range(0, slice_length, dtype=tf.dtypes.int64)
    random_slice = tf.gather(audio, slice_indices + random_offset, axis=0)
    return random_slice


def center_crop_signal(audio, slice_length):
    data_length = tf.shape(audio, out_type=tf.dtypes.int64)[0]
    if data_length == slice_length:
        return audio
    center_offset = data_length // 2
    slice_indices = tf.range(0, slice_length, dtype=tf.dtypes.int64)
    return tf.gather(audio, slice_indices + center_offset, axis=0)


def label_parser(example, mode="multiclass", num_classes=527):
    label = tf.sparse.to_dense(example['label'])
    if mode == "multilabel":
        example['label'] = tf.reduce_sum(tf.one_hot(label, num_classes, on_value=1., axis=-1), axis=0)
    else:
        example['label'] = tf.one_hot(label[0], num_classes, on_value=1.)
    return example


def contrastive_labels(example):
    labels = tf.range(0, example['anchor'].shape[0])
    labels = tf.one_hot(labels, example['anchor'].shape[0], on_value=1.)
    example['label'] = labels
    return example


def map_torch_batched_feature_extractor(example, feature_extractor):
    example['audio'] = tf.numpy_function(feature_extractor, [example['audio']], Tout=tf.float32)
    return example


def map_dtype(example, desired=tf.float32):
    example['audio'] = tf.cast(example['audio'], desired)
    example['label'] = tf.cast(example['label'], desired)
    return example


""" ConvNeXt
Paper: `A ConvNet for the 2020s` - https://arxiv.org/pdf/2201.03545.pdf

Adapted from the original PyTorch code (https://github.com/facebookresearch/ConvNeXt),
and Ross Wightman's implementation for timm (https://github.com/rwightman/pytorch-image-models)

Modifications and additions for audax by / Copyright 2022, Sarthak Yadav
"""
from functools import partial
import jax.numpy as jnp
from jax import lax, random
from flax import linen as nn
from .layers.drop import DropPath
from .layers.mlp import Mlp
from jax import dtypes
from typing import Optional, Tuple, Union, Callable, Any


def constant_init(key, shape, dtype=jnp.float_, constant=0.04):
    return jnp.ones(shape, dtypes.canonicalize_dtype(dtype)) * constant


def convnext_truncated_init(key, shape, dtype=jnp.float_, std=0.02):
    return random.truncated_normal(key, -2, 2, shape, dtype) * std


default_init = partial(convnext_truncated_init, std=0.02)


class ConvNeXtBlock(nn.Module):
    dim: int
    mlp_ratio: int = 4
    ls_init_value: float = 1e-6
    norm_layer: Callable = None
    dtype = jnp.float32
    drop_path: float = 0.

    @nn.compact
    def __call__(self, x, train: bool = True):
        shortcut = x
        x = nn.Conv(self.dim, kernel_size=(7, 7), padding="SAME",
                    feature_group_count=self.dim, dtype=self.dtype, kernel_init=default_init)(x)
        if self.norm_layer:
            x = self.norm_layer(dtype=self.dtype)(x)
        else:
            x = nn.LayerNorm(dtype=self.dtype)(x)
        x = Mlp(hidden_features=int(self.mlp_ratio*self.dim), out_features=self.dim,
                activation=nn.gelu, kernel_init=default_init)(x)

        if self.ls_init_value:
            gamma = self.param("gamma", partial(constant_init, constant=self.ls_init_value), [self.dim])
        else:
            gamma = None

        if gamma is not None:
            x = x * gamma.reshape((1,) * (x.ndim-1) + (-1,))
        
        if self.drop_path != 0.:
            x = DropPath(self.drop_path)(x, train=train) + shortcut
        else:
            x = x + shortcut
        return x


class ConvNeXtStage(nn.Module):
    out_channels: int
    stride: int = 2
    depth: int = 2
    dp_rates: Optional[float] = None
    ls_init_value: float = 1.0
    norm_layer: Callable = None
    cross_stage: Optional[float] = False
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, x, train: bool = True):
        in_chs = x.shape[-1]
        output = x
        if self.norm_layer is None:
            norm_layer = partial(nn.LayerNorm, dtype=self.dtype)
        else:
            norm_layer = partial(self.norm_layer, dtype=self.dtype)

        if in_chs != self.out_channels:
            output = norm_layer()(output)
            output = nn.Conv(self.out_channels, kernel_size=(self.stride, self.stride),
                             strides=(self.stride, self.stride), padding="VALID",
                             dtype=self.dtype, kernel_init=default_init)(output)

        dp_rates = self.dp_rates or [0.] * self.depth
        for j in range(self.depth):
            output = ConvNeXtBlock(dim=self.out_channels, drop_path=dp_rates[j],
                                   ls_init_value=self.ls_init_value,
                                   norm_layer=norm_layer)(output, train=train)
        return output


class ConvNeXt(nn.Module):
    r"""
    ConvNeXt
    A Jax impl of : `A ConvNet for the 2020s`  - https://arxiv.org/pdf/2201.03545.pdf

    Args:
        num_classes (int): Number of classes for classification head. Default: 1000
        depths (tuple(int)): Number of blocks at each stage. Default: [3, 3, 9, 3]
        dims (tuple(int)): Feature dimension at each stage. Default: [96, 192, 384, 768]
        drop_rate (float): Head dropout rate
        drop_path_rate (float): Stochastic depth rate. Default: 0.
        ls_init_value (float): Init value for Layer Scale. Default: 1e-6.
        head_init_scale (float): Init scaling value for classifier weights and biases. Default: 1.
    """
    num_classes: Optional[int] = 1000
    global_pool: str = "avg"
    depths: Optional[Tuple] = (3, 3, 9, 3)
    dims: Optional[Tuple] = (96, 192, 384, 768)
    output_stride: int = 32
    patch_size: int = 4
    ls_init_value: float = 1e-6
    head_init_scale: float = 1.
    norm_layer: Optional[Union[Callable, nn.Module]] = None
    drop_rate: float = 0.
    drop_path_rate: float = 0.
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, inputs, train: bool = True):
        assert self.output_stride == 32
        if self.norm_layer is None:
            norm_layer = partial(nn.LayerNorm, dtype=self.dtype)
        else:
            norm_layer = partial(self.norm_layer, dtype=self.dtype)

        # STEM
        outputs = nn.Conv(self.dims[0], kernel_size=(self.patch_size, self.patch_size),
                          strides=(self.patch_size, self.patch_size),
                          padding="VALID", dtype=self.dtype, kernel_init=default_init)(inputs)
        outputs = norm_layer()(outputs)
        # END STEM

        curr_stride = self.patch_size

        for i in range(4):
            stride = 2 if i > 0 else 1
            curr_stride *= stride
            out_chs = self.dims[i]
            outputs = ConvNeXtStage(
                out_channels=out_chs, stride=stride, depth=self.depths[i],
                ls_init_value=self.ls_init_value, norm_layer=norm_layer, dtype=self.dtype)(outputs, train=train)

        if self.global_pool == "avg":
            outputs = jnp.mean(outputs, axis=(1, 2))
        elif self.global_pool == "max":
            outputs = jnp.max(outputs, axis=(1, 2))
        else:
            raise ValueError("Unsupported pool type value : {}".format(self.global_pool))

        if self.num_classes is not None:
            outputs = norm_layer(name="fc_norm")(outputs)
            if self.drop_rate != 0.:
                outputs = nn.Dropout(self.drop_rate, deterministic=not train, name="fc_dropout")(outputs)
            outputs = nn.Dense(self.num_classes, name="fc", kernel_init=default_init)(outputs)

        return outputs


# drop_path_rate corresponds to Stochastic Depth parameter
# for ConvNeXT-T/S/B/L, Stochastic Depths corresponding to ImageNet-1k training are 0.1/0.4/0.5/0.5 resp.
# We don't use them by default

convnext_tiny = partial(ConvNeXt, depths=(3, 3, 9, 3), dims=(96, 192, 384, 768), drop_path_rate=0.1)
convnext_small = partial(ConvNeXt, depths=[3, 3, 27, 3], dims=[96, 192, 384, 768], drop_path_rate=0.4)
convnext_base = partial(ConvNeXt, depths=[3, 3, 27, 3], dims=[128, 256, 512, 1024])
convnext_large = partial(ConvNeXt, depths=[3, 3, 27, 3], dims=[192, 384, 768, 1536])
convnext_xlarge = partial(ConvNeXt, depths=[3, 3, 27, 3], dims=[256, 512, 1024, 2048])



"""
Implementation of the LEAF audio frontend based on the paper

[LEAF: A LEARNABLE FRONTEND FOR AUDIO CLASSIFICATION](https://openreview.net/forum?id=jM76BCb6F9m),
based on the official implementation here https://github.com/google-research/leaf-audio

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import flax.linen.initializers
import jax
from jax import dtypes
import math
from functools import partial
from typing import Union, Callable, Any, Iterable, Optional
from jax import numpy as jnp, lax
from flax import linen as nn
import warnings
from ..commons import utils
from ..core import functional


PRNGKey = Any
Shape = Iterable[int]
Dtype = Any  # this could be a real type?
Array = Any


def squared_modulus_activation(inputs):
    """
    Squared Modules Activation function as used in Leaf
    :param inputs:
    :return:
    """
    inputs = inputs.transpose((0, 2, 1))
    outputs = 2 * nn.avg_pool(inputs ** 2, (2,), strides=(2,))
    return outputs.transpose((0, 2, 1))


def gabor_impulse_response(t, center, fwhm):
    denominator = 1. / (jnp.sqrt(2.0 * math.pi) * fwhm)
    gaussian = jnp.exp(jnp.tensordot(1.0 / (2. * fwhm**2), -t**2, axes=0))
    center_frequency_complex = center.astype(jnp.complex64)
    t_complex = t.astype(jnp.complex64)
    sinusoid = jnp.exp(1j*jnp.tensordot(center_frequency_complex,t_complex,axes=0))
    denominator = denominator.astype(jnp.complex64)[:, jnp.newaxis]
    gaussian = gaussian.astype(jnp.complex64)
    return denominator * sinusoid * gaussian


def gabor_filters(kernel, size: int = 401):
    t = jnp.arange(-(size//2), (size+1)//2, dtype=jnp.float32)
    return gabor_impulse_response(t, center=kernel[:, 0], fwhm=kernel[:, 1])


def gaussian_lowpass_kernel(sigma, filter_size: int):
    sigma = jnp.clip(sigma, a_min=(2. / filter_size), a_max=0.5)
    t = jnp.arange(0, filter_size, dtype=jnp.float32)
    t = t.reshape((1, filter_size, 1, 1))
    numerator = t - 0.5 * (filter_size - 1)
    denominator = sigma * 0.5 * (filter_size - 1)
    return jnp.exp(-0.5 * (numerator / denominator) ** 2)


def constant_init(key, shape, dtype=jnp.float_, constant=0.04):
    return jnp.ones(shape, dtypes.canonicalize_dtype(dtype)) * constant


# TODO: Tidy up gabor_init and this mess, can be really simplified

class GaborFilterHelper:
    def __init__(self,
                 n_filters: int = 40,
                 min_freq: float = 0.,
                 max_freq: float = 8000.,
                 sample_rate: int = 16000,
                 window_len: int = 401,
                 n_fft: int = 512,
                 normalize_energy: bool = False
                 ):
        super(GaborFilterHelper, self).__init__()
        self.n_filters = n_filters
        self.min_freq = min_freq
        self.max_freq = max_freq
        self.sample_rate = sample_rate
        self.window_len = window_len
        self.n_fft = n_fft
        self.normalize_energy = normalize_energy
        mel_filters = functional.melscale_fbanks(n_freqs=(self.n_fft // 2) + 1, n_mels=self.n_filters,
                                                      sample_rate=self.sample_rate, f_min=self.min_freq,
                                                      f_max=self.max_freq, norm=None)
        mel_filters = mel_filters.transpose((1, 0))
        if self.normalize_energy:
            mel_filters = mel_filters / self._mel_filters_areas(mel_filters)
        self.mel_filters = mel_filters

    @property
    def gabor_params_from_mels(self):
        coeff = jnp.sqrt(2. * jnp.log(2.)) * self.n_fft
        sqrt_filters = jnp.sqrt(self.mel_filters)
        center_frequencies = jnp.argmax(sqrt_filters, axis=1)
        peaks = jnp.max(sqrt_filters, axis=1, keepdims=True)
        half_magnitudes = peaks / 2.
        fwhms = jnp.sum(
            (sqrt_filters >= half_magnitudes).astype(jnp.float32), axis=1)
        return jnp.stack([center_frequencies * 2 * jnp.pi / self.n_fft, coeff / (jnp.pi * fwhms)], axis=1)

    def _mel_filters_areas(self, filters):
        peaks = jnp.max(filters, axis=1, keepdims=True)
        return peaks * (jnp.sum((filters > 0).astype(jnp.float32), axis=1, keepdims=True) + 2) * jnp.pi / self.n_fft

    # @property
    # def mel_filters(self):
    #     """
    #     this is called only once at initialization
    #     """
    #     mel_filters = functional.melscale_fbanks(n_freqs=(self.n_fft // 2) + 1, n_mels=self.n_filters,
    #                                              sample_rate=self.sample_rate, f_min=self.min_freq,
    #                                              f_max=self.max_freq, norm=None)
    #     mel_filters = mel_filters.transpose((1, 0))
    #     if self.normalize_energy:
    #         mel_filters = mel_filters / self._mel_filters_areas(mel_filters)
    #     return mel_filters

    # def get_gabor_filters(self):
    #     gabor_params = self.gabor_params_from_mels()
    #     gabor_filters = gabor_filters(gabor_params, size=self.window_len)
    #     output = gabor_filters * jnp.sqrt(
    #         self._mel_filters_areas(self.mel_filters()) * 2 * math.sqrt(math.pi) * gabor_params[:, 1:2]
    #     ).astype(jnp.complex64)
    #     return output


class GaborInit:
    def __init__(self, default_window_len=401, **kwargs):
        super(GaborInit, self).__init__()
        self.def_win_len = default_window_len
        self._kwargs = kwargs

    def __call__(self, shape, dtype=None):
        n_filters = shape[0] if len(shape) == 2 else shape[-1] // 2
        window_len = self.def_win_len if len(shape) == 2 else shape[0]
        gabor_filters = GaborFilterHelper(n_filters=n_filters, window_len=window_len, **self._kwargs)
        if len(shape) == 2:
            return gabor_filters.gabor_params_from_mels
        else:
            # only needed in case of > 2-dim weights
            # even_indices = torch.arange(start=0, end=shape[2], step=2)
            # odd_indices = torch.arange(start=1, end=shape[2], step=2)
            # filters = gabor_filters.gabor_filters()
            raise NotImplementedError("implementation incomplete. Use even valued filter dimensions")


def gabor_init(key, shape, dtype=jnp.float_,
               window_len=401,
               sample_rate=16000, min_freq=60., max_freq=7800.,
               n_fft=512,
               normalize_energy: bool = False):
    init = GaborInit(default_window_len=window_len,
                     sample_rate=sample_rate, min_freq=min_freq, max_freq=max_freq,
                     n_fft=n_fft,
                     normalize_energy=normalize_energy)
    return init(shape, dtype=dtype)


class GaussianLowPassPooling(nn.Module):
    """
    Gaussian Lowpass pooling as found in original LEAF implementation
    """
    kernel_size: int
    stride: int
    in_channels: int
    padding: Union[str, int] = 'SAME'
    use_bias: bool = False
    dtype: Dtype = jnp.float32
    param_dtype: Dtype = jnp.float32
    precision: Any = None
    kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = None
    bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = nn.initializers.zeros

    @nn.compact
    def __call__(self, inputs):
        inputs = jnp.asarray(inputs, self.dtype)

        kernel_size = [self.kernel_size]

        def maybe_broadcast(x):
            if x is None:
                # backward compatibility with using None as sentinel for
                # broadcast 1
                x = 1
            if isinstance(x, int):
                return (x,) * len(kernel_size)
            return x

        is_single_input = False
        if inputs.ndim == len(kernel_size) + 1:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)

        kernel_shape = (1, 1, self.in_channels, 1)
        kernel = self.param('kernel', self.kernel_init, kernel_shape, self.param_dtype)
        kernel = jnp.asarray(kernel, self.dtype)

        kernel = gaussian_lowpass_kernel(kernel, self.kernel_size)
        kernel = kernel.reshape(-1, self.kernel_size, self.in_channels)
        kernel = kernel.transpose(1, 0, 2)  # kernel needs to be of this shape
        padding_lax = self.padding
        dimension_numbers = utils.conv_dimension_numbers(inputs.shape)
        lhs_dilation = maybe_broadcast(1)
        rhs_dilation = maybe_broadcast(1)
        strides = maybe_broadcast(self.stride)

        y = lax.conv_general_dilated(
            inputs,
            kernel,
            strides,
            padding_lax,
            lhs_dilation=lhs_dilation,
            rhs_dilation=rhs_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=self.in_channels,
            precision=self.precision)
        if self.use_bias:
            bias = self.param('bias', self.bias_init, (self.in_channels,), self.param_dtype)
            bias = jnp.asarray(bias, self.dtype)
            y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
        return y


class ExponentialMovingAverage(nn.Module):
    """
    ExponentialMovingAverage as used in LEAF implementation
    Usage:

    >>> init_func = partial(constant_init, constant=0.04)
    >>> exp = ExponentialMovingAverage(in_channels=40, per_channel=True, kernel_init=init_func)
    >>> # now initialize and use like regular linen Module

    """
    in_channels: int
    per_channel: bool = False
    trainable: bool = True
    dtype: Dtype = jnp.float32
    param_dtype: Dtype = jnp.float32
    kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = None

    @nn.compact
    def __call__(self, inputs):
        if self.kernel_init is None:
            raise ValueError("self.kernel_init is none.")
        inputs = jnp.asarray(inputs, self.dtype)
        kernel_shape = (self.in_channels,) if self.per_channel else (1,)
        if self.trainable:
            kernel = self.param('smooth', self.kernel_init, kernel_shape, self.param_dtype)
        else:
            kernel = self.kernel_init(0, kernel_shape, dtype=self.param_dtype)

        kernel = jnp.clip(kernel, a_min=0., a_max=1.)
        initial_state = inputs[:, 0, :]

        def func(a, x):
            a = kernel * x + (1.0 - kernel) * a
            return a, a

        _, result = lax.scan(func, initial_state, inputs.transpose((1, 0, 2)))
        return result.transpose((1, 0, 2))


class PCENLayer(nn.Module):
    in_channels: int
    alpha_coef: float = 0.96
    smooth_coef: float = 0.04
    delta_coef: float = 2.0
    root_coef: float = 2.0
    floor_coef: float = 1e-6
    trainable: bool = True
    learn_smooth_coef: bool = True
    per_channel_smooth_coef: bool = False
    dtype: Dtype = jnp.float32
    param_dtype: Dtype = jnp.float32

    @nn.compact
    def __call__(self, inputs) -> Any:
        inputs = inputs.astype(self.dtype)
        assert self.learn_smooth_coef, "Only sPCEN is supported"

        if self.trainable:
            alpha = self.param("alpha", partial(constant_init, constant=self.alpha_coef), [self.in_channels],
                               self.param_dtype)
            delta = self.param("delta", partial(constant_init, constant=self.delta_coef), [self.in_channels],
                               self.param_dtype)
            root = self.param("root", partial(constant_init, constant=self.root_coef), [self.in_channels],
                              self.param_dtype)
        else:
            alpha = constant_init(0, [self.in_channels], constant=self.alpha_coef, dtype=self.param_dtype)
            delta = constant_init(0, [self.in_channels], constant=self.delta_coef, dtype=self.param_dtype)
            root = constant_init(0, [self.in_channels], constant=self.root_coef, dtype=self.param_dtype)

        alpha = jnp.minimum(alpha, 1.0)
        root = jnp.maximum(root, 1.0)

        ema = ExponentialMovingAverage(self.in_channels, self.per_channel_smooth_coef,
                                       trainable=self.trainable, dtype=self.dtype, param_dtype=self.param_dtype,
                                       kernel_init=partial(constant_init, constant=self.smooth_coef))

        ema_smoother = ema(inputs)
        one_over_root = 1. / root
        output = ((inputs / (
                    self.floor_coef + ema_smoother) ** alpha + delta) ** one_over_root - delta ** one_over_root)
        output = output.astype(self.dtype)
        return output


class GaborConstraint:
    def __init__(self, kernel_size: int):
        self.kernel_size = kernel_size

    def __call__(self, kernel):
        mu_lower = 0.
        mu_upper = math.pi
        sigma_lower = 4 * math.sqrt(2 * math.log(2)) / math.pi
        sigma_upper = self.kernel_size * math.sqrt(2 * math.log(2)) / math.pi
        clipped_mu = jnp.clip(kernel[:, 0], mu_lower, mu_upper)
        clipped_sigma = jnp.clip(kernel[:, 1], sigma_lower, sigma_upper)
        return jnp.stack([clipped_mu, clipped_sigma], axis=1)


class GaborConv1D(nn.Module):
    filters: int
    kernel_size: int
    stride: int
    padding: str = "SAME"
    use_bias: bool = False
    sort_filter: bool = False
    constraint: Optional[Callable] = None
    kernel_init: Optional[Callable[[PRNGKey, Shape, Dtype], Array]] = None
    dtype: Any = jnp.float32
    param_dtype: Any = jnp.float32
    precision: Any = None

    @nn.compact
    def __call__(self, inputs):
        inputs = jnp.asarray(inputs, self.dtype)
        kernel_size = [self.kernel_size]
        num_filters = self.filters // 2

        if not self.kernel_init:
            raise ValueError("Initialized without kernel_init argument")

        def maybe_broadcast(x):
            if x is None:
                # backward compatibility with using None as sentinel for
                # broadcast 1
                x = 1
            if isinstance(x, int):
                return (x,) * len(kernel_size)
            return x

        kernel = self.param('gabor_kernel', self.kernel_init, [num_filters, 2], self.param_dtype)
        if self.constraint is None:
            warnings.warn("GaborConv1d was initialized without a constraint..")
        else:
            kernel = self.constraint(kernel)
        if self.sort_filter:
            # TODO add sort filter functionality
            raise NotImplementedError("filter sorting not implemented yet")
        filters = gabor_filters(kernel, self.kernel_size)
        real_filters = jnp.real(filters)
        imag_filters = jnp.imag(filters)
        stacked_filters = jnp.stack([real_filters, imag_filters], axis=1)
        stacked_filters = stacked_filters.reshape((2 * num_filters, self.kernel_size))
        stacked_filters = jnp.expand_dims(stacked_filters.transpose((1, 0)), axis=1)

        padding_lax = self.padding
        dimension_numbers = utils.conv_dimension_numbers(inputs.shape)
        lhs_dilation = maybe_broadcast(1)
        rhs_dilation = maybe_broadcast(1)
        stride = maybe_broadcast(self.stride)
        y = lax.conv_general_dilated(
            inputs,
            stacked_filters,
            stride,
            padding_lax,
            lhs_dilation=lhs_dilation,
            rhs_dilation=rhs_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=1,
            precision=self.precision)

        if self.use_bias:
            bias = self.param('gabor_bias', self.bias_init, (num_filters * 2,), self.param_dtype)
            bias = jnp.asarray(bias, self.dtype)
            y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
        return y


class Leaf(nn.Module):
    n_filters: int = 40
    sample_rate: int = 16000
    window_len: float = 25.
    window_stride: float = 10.
    min_freq: float = 60.
    max_freq: float = 7800.
    complex_conv_init: Optional[Union[Callable, str]] = "default"
    train_pcen: Optional[bool] = True
    pooling_param = 0.4
    use_bias: bool = False
    spec_augment: bool = False
    dtype: Dtype = jnp.float32
    param_dtype: Dtype = jnp.float32
    precision: Any = None

    def setup(self):
        window_size = int(self.sample_rate * self.window_len // 1000 + 1)
        window_stride = int(self.sample_rate * self.window_stride // 1000)
        if type(self.complex_conv_init) == str:
            if self.complex_conv_init.lower() == "default":
                conv_init_func = partial(gabor_init,
                                         window_len=window_size,
                                         sample_rate=self.sample_rate,
                                         min_freq=self.min_freq,
                                         max_freq=self.max_freq,
                                         n_fft=512,
                                         normalize_energy=False)
            elif self.complex_conv_init.lower() == "randn":
                conv_init_func = flax.linen.initializers.normal()
            elif self.complex_conv_init.lower() == "lecun_normal":
                conv_init_func = flax.linen.initializers.lecun_normal()
            elif self.complex_conv_init.lower() == "kaiming_normal":
                conv_init_func = flax.linen.initializers.kaiming_normal()
            else:
                raise ValueError("Unsupported 'str' valued argument {}, should be one of [randn, lecun_normal, kaiming_normal]. Pass a callable instead")
        else:
            conv_init_func = self.complex_conv_init

        self.complex_conv = GaborConv1D(filters=self.n_filters*2,
                                        kernel_size=window_size, stride=1,
                                        use_bias=self.use_bias,
                                        constraint=GaborConstraint(window_size),
                                        kernel_init=conv_init_func)

        self.pooling = GaussianLowPassPooling(kernel_size=window_size, stride=window_stride,
                                              in_channels=self.n_filters, use_bias=False, dtype=self.dtype,
                                              param_dtype=self.param_dtype, precision=self.precision,
                                              kernel_init=partial(constant_init,
                                                                   constant=self.pooling_param))
        self.compression = PCENLayer(in_channels=self.n_filters, alpha_coef=0.96,
                                     smooth_coef=0.04, delta_coef=2.0, floor_coef=1e-12,
                                     trainable=self.train_pcen, learn_smooth_coef=True,
                                     per_channel_smooth_coef=True, dtype=self.dtype)
        if self.spec_augment:
            raise NotImplementedError("spec augment functionality not yet added.")

    def __call__(self, inputs):
        outputs = inputs[:, :, jnp.newaxis] if inputs.ndim < 3 else inputs
        outputs = outputs.astype(self.dtype)
        outputs = self.complex_conv(outputs)
        outputs = squared_modulus_activation(outputs)
        outputs = self.pooling(outputs)
        outputs = jnp.maximum(outputs, 1e-5)
        outputs = self.compression(outputs)
        outputs = outputs.astype(self.dtype)
        return outputs



"""
Implementation of the SincNet audio frontend based on the paper

Ravanelli, Mirco, and Yoshua Bengio. "Speaker recognition from raw waveform with sincnet."
In 2018 IEEE Spoken Language Technology Workshop (SLT), pp. 1021-1028. IEEE, 2018.

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import jax
import math
from functools import partial
from typing import Union, Callable, Any, Iterable, Optional
from jax import numpy as jnp, lax
from flax import linen as nn
from ..commons import utils

PRNGKey = Any
Shape = Iterable[int]
Dtype = Any  # this could be a real type?
Array = Any


def sinc_impulse_response(t: jnp.array, frequency: jnp.array) -> jnp.array:
    """Computes the sinc impulse response."""
    return jnp.sin(2 * math.pi * frequency * t) / (2 * math.pi * frequency * t)


def sinc_filters(cutoff_freq_low: jnp.array,
                 cutoff_freq_high: jnp.array,
                 size: int = 401,
                 sample_rate: int = 16000) -> jnp.array:
    """Computes the sinc filters from its parameters for a given size.

    Sinc is not defined in zero so we need to separately compute negative
    (left_range) and positive part (right_range).

    Args:
        cutoff_freq_low: tf.Tensor<float>[1, filters] the lower cutoff frequencies
        of the bandpass.
        cutoff_freq_high: tf.Tensor<float>[1, filters] the upper cutoff frequencies
        of the bandpass.
        size: the size of the output tensor.
        sample_rate: audio sampling rate

    Returns:
        A jnp.tensor<float>[size, filters].
    """
    left_range = jnp.arange(-(size // 2), 0, dtype=jnp.float32)[:, jnp.newaxis] / float(sample_rate)
    right_range = jnp.arange(1, (size // 2) + 1, dtype=jnp.float32)[:, jnp.newaxis] / float(sample_rate)

    high_pass_left_range = 2 * cutoff_freq_high * sinc_impulse_response(left_range, cutoff_freq_high)
    high_pass_right_range = 2 * cutoff_freq_high * sinc_impulse_response(right_range, cutoff_freq_high)

    low_pass_left_range = 2 * cutoff_freq_low * sinc_impulse_response(left_range, cutoff_freq_low)
    low_pass_right_range = 2 * cutoff_freq_low * sinc_impulse_response(right_range, cutoff_freq_low)

    high_pass = jnp.concatenate([high_pass_left_range, 2 * cutoff_freq_high, high_pass_right_range], axis=0)
    low_pass = jnp.concatenate([low_pass_left_range, 2 * cutoff_freq_low, low_pass_right_range], axis=0)

    band_pass = high_pass - low_pass
    return band_pass / jnp.max(band_pass, axis=0, keepdims=True)


def sinc_init(key, shape, dtype=jnp.float_,
              sample_rate=16000, min_low_hz=50., min_band_hz=50.,
              sampling="default"):
    init = SincInit(sample_rate=sample_rate, min_low_hz=min_low_hz, 
                    min_band_hz=min_band_hz, sampling=sampling)
    return init(key, shape, dtype=dtype)


class SincInit:
    def __init__(self,
                 sample_rate: int = 16000,
                 min_low_hz: float = 50.,
                 min_band_hz: float = 50.,
                 sampling="default"):
        self._sample_rate = sample_rate
        self._min_low_hz = min_low_hz
        self._min_band_hz = min_band_hz
        self._sampling = sampling

    def __call__(self, key, shape, dtype=jnp.float32):
        filters = shape[0]
        low_hz = self._min_low_hz
        high_hz = 0.5 * self._sample_rate - (low_hz + self._min_band_hz)
        low_mel = utils.hz2mel(low_hz)
        high_mel = utils.hz2mel(high_hz)
        bandpass_mel = jnp.linspace(low_mel, high_mel, filters + 1)
        bandpass_hz = utils.mel2hz(bandpass_mel)
        left_edge = bandpass_hz[:-1]
        filter_width = bandpass_hz[1:] - bandpass_hz[:-1]
        return jnp.stack([left_edge, filter_width], axis=1).astype(dtype)


class SincConv1D(nn.Module):
    filters: int
    kernel_size: int
    stride: int
    kernel_init: Callable[[PRNGKey, Shape, Dtype], Array]
    window_fn: Optional[Callable] = jnp.hamming
    trainable: bool = True
    padding: str = "SAME"
    sample_rate: int = 16000
    min_low_hz: float = 50.
    min_band_hz: float = 50.
    use_bias: bool = False
    dtype: Any = jnp.float32
    param_dtype: Any = jnp.float32
    precision: Any = None

    @nn.compact
    def __call__(self, inputs):
        inputs = jnp.asarray(inputs, self.dtype)
        kernel_size = [self.kernel_size]
        num_filters = self.filters // 2

        if not self.kernel_init:
            raise ValueError("initialized without kernel_init")

        # if not self.window:
        #     raise ValueError("initialized without a window")

        # if len(self.window) != self.kernel_size:
        #     raise ValueError("window size: {} is not equal to kernel_size: {}".format(len(self.window), kernel_size))

        def maybe_broadcast(x):
            if x is None:
                # backward compatibility with using None as sentinel for
                # broadcast 1
                x = 1
            if isinstance(x, int):
                return (x,) * len(kernel_size)
            return x

        kernel = self.param('sinc_kernel', self.kernel_init, [num_filters, 2], self.param_dtype)
        window = self.window_fn(self.kernel_size)
        left_edge = self.min_low_hz + jnp.abs(kernel[:, 0])
        right_edge = jnp.clip(
            left_edge + self.min_band_hz + jnp.abs(kernel[:, 1]),
            a_min=self.min_low_hz,
            a_max=self.sample_rate / 2
        )
        filters = sinc_filters(
            left_edge[jnp.newaxis, :], right_edge[jnp.newaxis, :],
            self.kernel_size) * window[:, jnp.newaxis]
        filters = jnp.expand_dims(filters, axis=1)
        padding_lax = self.padding
        dimension_numbers = utils.conv_dimension_numbers(inputs.shape)
        lhs_dilation = maybe_broadcast(1)
        rhs_dilation = maybe_broadcast(1)
        stride = maybe_broadcast(self.stride)

        y = lax.conv_general_dilated(
            inputs,
            filters,
            stride,
            padding_lax,
            lhs_dilation=lhs_dilation,
            rhs_dilation=rhs_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=1,
            precision=self.precision)

        if self.use_bias:
            bias = self.param('sinc_bias', self.bias_init, (num_filters,), self.param_dtype)
            bias = jnp.asarray(bias, self.dtype)
            y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
        return y


class SincNet(nn.Module):
    n_filters: int = 40
    sample_rate: int = 16000
    window_len: float = 25.
    window_stride: float = 10.
    min_low_hz = 50.
    min_band_hz = 50.
    pooling_param = 0.4
    use_bias: bool = False
    spec_augment: bool = False
    dtype: Dtype = jnp.float32
    param_dtype: Dtype = jnp.float32
    complex_conv_init: Optional[Union[Callable, str]] = "default"
    precision: Any = None

    def setup(self):
        window_size = int(self.sample_rate * self.window_len // 1000 + 1)
        window_stride = int(self.sample_rate * self.window_stride // 1000)
        self.complex_conv = SincConv1D(filters=self.n_filters * 2,
                                       kernel_size=window_size, stride=1,
                                       use_bias=self.use_bias,
                                       sample_rate=self.sample_rate,
                                       kernel_init=partial(sinc_init,
                                                           sample_rate=self.sample_rate,
                                                           min_low_hz=self.min_low_hz,
                                                           min_band_hz=self.min_band_hz,
                                                           sampling=self.complex_conv_init))

        self.pooling = partial(nn.max_pool, window_shape=(window_size,), strides=(window_stride,), padding="SAME")
        self.compression = nn.LayerNorm()

        if self.spec_augment:
            raise NotImplementedError("spec augment functionality not yet added.")

    def __call__(self, inputs):
        outputs = inputs[:, :, jnp.newaxis] if inputs.ndim < 3 else inputs
        outputs = outputs.astype(self.dtype)
        outputs = self.complex_conv(outputs)
        outputs = nn.leaky_relu(outputs, negative_slope=0.2)
        outputs = self.pooling(outputs)
        outputs = jnp.maximum(outputs, 1e-5)
        outputs = self.compression(outputs)
        outputs = outputs.astype(self.dtype)
        return outputs



"""
Modified training state with easy param freezing support and updating rngs over training iterations

Written for audax by / Copyright 2022, Sarthak Yadav
"""
from typing import Any, Callable, Dict
import jax
from flax import core
from flax import struct
import optax
import flax


class TrainState_v2(struct.PyTreeNode):
    """Simple train state for the common case with a single Optax optimizer.

      Synopsis::
          state = TrainState.create(
              apply_fn=model.apply,
              params=variables['params'],
              tx=tx)
          grad_fn = jax.grad(make_loss_fn(state.apply_fn))
          for batch in data:
            grads = grad_fn(state.params, batch)
            state = state.apply_gradients(grads=grads)

      Note that you can easily extend this dataclass by subclassing it for storing
      additional data (e.g. additional variable collections).

      For more exotic usecases (e.g. multiple optimizers) it's probably best to
      fork the class and modify it.

      Args:
        step: Counter starts at 0 and is incremented by every call to
          `.apply_gradients()`.
        apply_fn: Usually set to `model.apply()`. Kept in this dataclass for
          convenience to have a shorter params list for the `train_step()` function
          in your training loop.
        params: The parameters to be updated by `tx` and used by `apply_fn`.
        frozen_params:
        tx: An Optax gradient transformation.
        opt_state: The state for `tx`.
    """
    step: int
    batch_stats: Any
    dynamic_scale: flax.optim.DynamicScale
    params: core.FrozenDict[str, Any]
    frozen_params: core.FrozenDict[str, Any]
    aux_rng_keys: core.FrozenDict[str, Any]
    opt_state: optax.OptState
    apply_fn: Callable = struct.field(pytree_node=False)
    tx: optax.GradientTransformation = struct.field(pytree_node=False)

    def apply_gradients(self, *, grads, **kwargs):
        """Updates `step`, `params`, `opt_state` and `**kwargs` in return value.

        Note that internally this function calls `.tx.update()` followed by a call
        to `optax.apply_updates()` to update `params` and `opt_state`.

        Args:
          grads: Gradients that have the same pytree structure as `.params`.
          **kwargs: Additional dataclass attributes that should be `.replace()`-ed.

        Returns:
          An updated instance of `self` with `step` incremented by one, `params`
          and `opt_state` updated by applying `grads`, and additional attributes
          replaced as specified by `kwargs`.
        """
        updates, new_opt_state = self.tx.update(
            grads, self.opt_state, self.params)
        new_params = optax.apply_updates(self.params, updates)

        rng_keys = self.update_rng_keys()

        return self.replace(
            step=self.step + 1,
            params=new_params,
            frozen_params=self.frozen_params,
            opt_state=new_opt_state,
            aux_rng_keys=rng_keys,
            **kwargs)

    def update_rng_keys(self):
        unfrozen = flax.core.unfreeze(self.aux_rng_keys)
        for k in self.aux_rng_keys.keys():
            unfrozen[k] = jax.random.split(unfrozen[k], 1)[0]
        return flax.core.freeze(unfrozen)

    @property
    def get_all_params(self):
        return {**self.params, **self.frozen_params}

    @classmethod
    def create(cls, *, apply_fn, params, frozen_params, tx, aux_rng_keys, **kwargs):
        """Creates a new instance with `step=0` and initialized `opt_state`."""
        opt_state = tx.init(params)
        return cls(
            step=0,
            apply_fn=apply_fn,
            params=params,
            frozen_params=frozen_params,
            tx=tx,
            opt_state=opt_state,
            aux_rng_keys=aux_rng_keys,
            **kwargs,
        )



"""
Supervised training boilerplate code used for training a audio classifier in the provided recipes

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import functools
import os
import time
from typing import Any

from absl import logging
from clu import metric_writers
from clu import periodic_actions
import flax
from flax import jax_utils
from flax import optim
from flax.training import checkpoints
from flax.training import common_utils
from flax.training import train_state
import jax
from jax import lax
import jax.numpy as jnp
from jax import random
import ml_collections
import optax
import tensorflow as tf
from sklearn.metrics import average_precision_score
from ..transforms import mixup


try:
    # Disable all GPUS
    tf.config.set_visible_devices([], 'GPU')
    visible_devices = tf.config.get_visible_devices()
    for device in visible_devices:
        assert device.device_type != 'GPU'
except:
    # Invalid device or cannot modify virtual devices once initialized.
    pass

from .misc import TrainingMode, Features, DataSplit
from . import training_utilities
from .data_v2.helpers import prepare_datasets_v2
from .. import models
# from encoders.classifier import Classifier
import wandb
from audax import frontends


def create_model(*, model_cls, half_precision, num_classes, 
                 drop_rate=None, spec_aug=None, **kwargs):
    platform = jax.local_devices()[0].platform
    if half_precision:
        if platform == 'tpu':
            model_dtype = jnp.bfloat16
        else:
            model_dtype = jnp.float16
    else:
        model_dtype = jnp.float32
    return models.Classifier(model_cls=model_cls, num_classes=num_classes,
                             dtype=model_dtype, drop_rate=drop_rate, spec_aug=spec_aug, 
                             **kwargs)


def train_step(state, batch, learning_rate_fn,
               cost_func,
               mode=TrainingMode.MULTICLASS,
               mixup_func=None,
               mixup_criterion_func=None):
    """Perform a single training step."""
    inputs = batch['audio']
    targets = batch["label"]
    if mixup_func is not None:
        inputs, y_a, y_b, lam = mixup_func(state.aux_rng_keys["mixup"], inputs, targets)

    def loss_fn(params):
        logits, new_model_state = state.apply_fn(
            {
                'params': {**params, **state.frozen_params},        # we don't use state.get_all_params here
                'batch_stats': state.batch_stats                    # since we'll override gradient computation
            },
            inputs,
            mutable=['batch_stats'],
            rngs=state.aux_rng_keys
        )
        if mixup_func is not None:
            loss = mixup_criterion_func(cost_func, logits, y_a, y_b, lam)
        else:
            loss = cost_func(logits, batch['label'])
        return loss, (new_model_state, logits)

    step = state.step
    dynamic_scale = state.dynamic_scale
    lr = learning_rate_fn(step)

    if dynamic_scale:
        grad_fn = dynamic_scale.value_and_grad(
            loss_fn, has_aux=True, axis_name='batch')
        dynamic_scale, is_fin, aux, grads = grad_fn(state.params)
        # dynamic loss takes care of averaging gradients across replicas
    else:
        grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
        aux, grads = grad_fn(state.params)
        # Re-use same axis_name as in the call to `pmap(...train_step...)` below.
        grads = lax.pmean(grads, axis_name='batch')

    new_model_state, logits = aux[1]
    metrics = training_utilities.compute_metrics(logits, batch['label'], mode=mode, cost_fn=cost_func)
    metrics['learning_rate'] = lr
    new_state = state.apply_gradients(
        grads=grads, batch_stats=new_model_state['batch_stats'])
    if dynamic_scale:
        new_state = new_state.replace(
            opt_state=jax.tree_multimap(
                functools.partial(jnp.where, is_fin),
                new_state.opt_state,
                state.opt_state),
            params=jax.tree_multimap(
                functools.partial(jnp.where, is_fin),
                new_state.params,
                state.params))
        metrics['scale'] = dynamic_scale.scale
    return new_state, metrics


def eval_step(state, batch, cost_func, mode=TrainingMode.MULTICLASS):
    variables = {
        'params': state.get_all_params,                    # absolutely ok to just use state.get_all_params here
        'batch_stats': state.batch_stats
    }
    logits = state.apply_fn(
        variables, batch['audio'], train=False, mutable=False)
    metrics = training_utilities.compute_metrics(logits, batch['label'], mode=mode, cost_fn=cost_func)
    return metrics, logits, batch['label']


def train_and_evaluate(config: ml_collections.ConfigDict,
                       workdir: str,
                       no_wandb: bool,
                       seed: int = 0):
    """Execute model training and evaluation loop.

      Args:
        config: Hyperparameter configuration for training and evaluation.
        workdir: Directory where the tensorboard summaries are written to.

      Returns:
        Final TrainState.
      """
    wandb_logger = None
    if not no_wandb:
        wandb_logger = wandb.init(project='{}'.format(config.wandb.get("project", "audax-supervised")),
                                  group="{}".format(config.data.dataset_name),
                                  config=config.to_dict(), name=workdir.split("/")[-1])
    writer = metric_writers.create_default_writer(
        logdir=workdir, just_logging=jax.process_index() != 0)
    # write config to workdir
    training_utilities.write_config_to_json(workdir, config)
    rng = random.PRNGKey(seed)
    if config.batch_size % jax.device_count() > 0:
        raise ValueError('Batch size must be divisible by the number of devices')
    local_batch_size = config.batch_size // jax.process_count()
    logging.info("Process count: {}".format(jax.process_count()))
    device = config.get("device", None)
    if device:
        devices = [jax.local_devices()[device]]
    else:
        devices = jax.local_devices()
    platform = devices[0].platform
    if config.half_precision:
        if platform == 'tpu':
            input_dtype = tf.bfloat16
        else:
            input_dtype = tf.float16
    else:
        input_dtype = tf.float32
    mode = TrainingMode(config.model.type)
    logging.info('Training mode is {}...'.format(mode))
    train_iter, eval_iter = prepare_datasets_v2(config, local_batch_size, input_dtype=input_dtype)
    train_iter = training_utilities.create_input_iter(train_iter, devices=devices)
    eval_iter = training_utilities.create_input_iter(eval_iter, devices=devices)

    if config.data.jax_transforms:
        tfs = training_utilities.get_feature_functions(config)
        if len(tfs) != 0:
            p_feature_extract_fn = jax.pmap(
                functools.partial(
                    training_utilities.apply_audio_transforms, transforms=tfs, 
                    dtype=training_utilities.get_dtype(config.half_precision),
                    normalize=config.data.get("normalize_batch", False)
                ), axis_name='batch', devices=devices)
        else:
            p_feature_extract_fn = None
    else:
        p_feature_extract_fn = None

    in_model_spec_aug = config.get("in_model_spec_aug", False)
    p_spec_aug, spec_aug_rng = training_utilities.get_spec_augment(config, devices, pmapped=not in_model_spec_aug)
    if p_spec_aug:
        logging.info("Using spec augment")
    num_examples = config.data.tr_samples
    steps_per_epoch = (num_examples // config.batch_size)

    if config.num_train_steps == -1:
        num_steps = int(steps_per_epoch * config.num_epochs)
    else:
        num_steps = config.num_train_steps

    if config.steps_per_eval == -1:
        num_validation_examples = config.data.eval_samples
        steps_per_eval = num_validation_examples // config.batch_size
    else:
        steps_per_eval = config.steps_per_eval

    steps_per_checkpoint = steps_per_epoch
    base_learning_rate = config.opt.learning_rate * config.batch_size / 256.

    model_cls, frontend_cls = training_utilities.get_model_frontend_cls(config)
    model = create_model(
        model_cls=model_cls, half_precision=config.half_precision,
        frontend_cls=frontend_cls,
        num_classes=config.model.num_classes,
        spec_aug=p_spec_aug if in_model_spec_aug else None,
        drop_rate=config.model.get("fc_drop_rate", 0.))
    print(model)
    learning_rate_fn = training_utilities.create_learning_rate_fn(
        config, base_learning_rate, steps_per_epoch)

    if config.model.get("pretrained", None):
        state = training_utilities.create_train_state_from_pretrained(rng, config,
                                                                      model, learning_rate_fn,
                                                                      config.model.pretrained,
                                                                      config.model.get("pretrained_prefix",
                                                                                       "checkpoint_"),
                                                                      to_copy=['encoder'],
                                                                      fc_only=config.model.get("pretrained_fc_only",
                                                                                               False))
    else:
        state = training_utilities.create_train_state(rng, config, model, learning_rate_fn)
    state = training_utilities.restore_checkpoint(state, workdir)

    step_offset = int(state.step)
    state = jax_utils.replicate(state, devices=devices)

    label_smoothing_factor = config.opt.get("label_smoothing_factor", None)
    if label_smoothing_factor:
        logging.info("Training with Label Smoothing, alpha = {}".format(label_smoothing_factor))

    # setup funcs!
    if mode == TrainingMode.MULTICLASS:
        cost_fn = functools.partial(training_utilities.cross_entropy_loss, smoothing_factor=label_smoothing_factor)
    else:
        cost_fn = functools.partial(training_utilities.binary_xentropy_loss,  smoothing_factor=label_smoothing_factor)

    mixup_alpha = config.opt.get("mixup_alpha", 0.0)
    if mixup_alpha != 0.0 and (mode == TrainingMode.MULTILABEL or mode==TrainingMode.MULTICLASS):
        logging.info(f"Mixup with alpha {mixup_alpha} will be used for training")
        mixup_func = functools.partial(mixup.do_mixup, alpha=mixup_alpha, mode=mode)
        mixup_criterion_func = functools.partial(mixup.get_mixedup_xentropy, mode=mode)
    else:
        mixup_func = None
        mixup_criterion_func = None

    p_train_step = jax.pmap(
        functools.partial(train_step, learning_rate_fn=learning_rate_fn,
                          cost_func=cost_fn,
                          mode=mode, mixup_func=mixup_func, 
                          mixup_criterion_func=mixup_criterion_func),
        axis_name='batch', devices=devices)
    p_eval_step = jax.pmap(functools.partial(eval_step, mode=mode, cost_func=cost_fn),
                           axis_name='batch', devices=devices)

    train_metrics = []
    hooks = []
    if jax.process_index() == 0:
        hooks += [periodic_actions.Profile(num_profile_steps=5, logdir=workdir)]
    train_metrics_last_t = time.time()
    best_val_acc = 0.
    logging.info('Initial compilation, this might take some minutes...')

    for step, batch in zip(range(step_offset, num_steps), train_iter):
        if p_feature_extract_fn:
            batch['audio'] = p_feature_extract_fn(batch['audio'])
            if p_spec_aug and not in_model_spec_aug:
                batch['audio'], spec_aug_rng = p_spec_aug(batch['audio'], spec_aug_rng)
        if step == 0:
            print(batch['audio'].shape, batch['audio'].dtype)
            # jnp.save("test_{:05d}.npy".format(step), jax.device_get(batch['audio']))

        state, metrics = p_train_step(state, batch)
        for h in hooks:
            h(step)
        if step == step_offset:
            logging.info('Initial compilation completed.')

        if config.get('log_every_steps'):
            train_metrics.append(metrics)
            if (step + 1) % config.log_every_steps == 0:
                train_metrics = common_utils.get_metrics(train_metrics)
                summary = {
                    f'train_{k}': v
                    for k, v in jax.tree_map(lambda x: x.mean(), train_metrics).items()
                }
                summary['steps_per_second'] = config.log_every_steps / (
                        time.time() - train_metrics_last_t)
                writer.write_scalars(step + 1, summary)
                if wandb_logger:
                    wandb_logger.log(summary, step+1)
                train_metrics = []
                train_metrics_last_t = time.time()

        if (step + 1) % steps_per_epoch == 0:
            epoch = step // steps_per_epoch
            eval_metrics = []
            eval_logits = []
            eval_labels = []
            # sync batch statistics across replicas
            state = training_utilities.sync_batch_stats(state)

            for _ in range(steps_per_eval):
                eval_batch = next(eval_iter)
                if p_feature_extract_fn:
                    eval_batch['audio'] = p_feature_extract_fn(eval_batch['audio'])
                metrics, logits, labels = p_eval_step(state, eval_batch)
                eval_metrics.append(metrics)
                eval_logits.append(logits)
                eval_labels.append(labels)
                # print("logits.shape: {} | type: {}".format(logits.shape, type(logits)))

            eval_metrics = common_utils.get_metrics(eval_metrics)

            summary = jax.tree_map(lambda x: x.mean(), eval_metrics)
            if mode == TrainingMode.MULTILABEL:
                eval_logits = jnp.concatenate([jax.device_get(x) for x in eval_logits])
                eval_labels = jnp.concatenate([jax.device_get(x) for x in eval_labels])
                eval_logits = eval_logits.reshape(-1, eval_logits.shape[-1])
                eval_labels = eval_labels.reshape(-1, eval_labels.shape[-1])
                map_value = average_precision_score(eval_labels.astype('float32'), eval_logits.astype('float32'),
                                                    average="macro")
                summary['accuracy'] = map_value

            logging.info('eval epoch: %d, loss: %.4f, accuracy: %.4f',
                         epoch, summary['loss'], summary['accuracy'] * 100)

            if summary['accuracy'] >= best_val_acc:
                best_val_acc = summary['accuracy']
                state = training_utilities.sync_batch_stats(state)
                training_utilities.save_best_checkpoint(state, workdir, best_val_acc)

            writer.write_scalars(
                step + 1, {f'eval_{key}': val for key, val in summary.items()})
            writer.flush()
            if wandb_logger:
                wandb_logger.log({f'eval_{key}': val for key, val in summary.items()}, step + 1)

        if (step + 1) % steps_per_checkpoint == 0 or step + 1 == num_steps:
            state = training_utilities.sync_batch_stats(state)
            training_utilities.save_checkpoint(state, workdir)

    # Wait until computations are done before exiting
    jax.random.normal(jax.random.PRNGKey(0), ()).block_until_ready()
    if wandb_logger:
        wandb_logger.finish()
    return state



"""
helper functions for EfficientNet CNN models

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import jax
from jax import numpy as jnp
from flax import linen as nn
from typing import Callable, Any, Optional
from functools import partial


class SqueezeAndExcitation(nn.Module):
    """
    Squeeze-and-Excitation block based on reference implementation in torchvision
    https://github.com/pytorch/vision/blob/4bf6c6e4b4cf2dea9c7f9952aa30d5820ab8ae33/torchvision/ops/misc.py
    """
    input_channels: int
    squeeze_channels: int
    activation: Callable = nn.relu
    scale_activation: Callable = nn.sigmoid
    kernel_init: Callable = nn.initializers.kaiming_normal()
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, x):
        conv = partial(nn.Conv, kernel_size=(1, 1), padding="VALID",
                       use_bias=False, dtype=self.dtype, kernel_init=self.kernel_init)

        # AdaptiveAvgPool
        scale = jnp.asarray(x, jnp.float32)
        scale = scale.mean((1, 2), keepdims=True)
        scale = jnp.asarray(scale, self.dtype)

        scale = conv(self.squeeze_channels, name="reduce")(scale)
        scale = self.activation(scale)
        scale = conv(self.input_channels, name="expand")(scale)
        scale = self.scale_activation(scale)
        return x * scale


class ConvNormActivation(nn.Module):
    out_channels: int
    kernel_size: int = 3
    stride: int = 1
    padding: Optional[int] = None
    groups: int = 1
    norm_layer: Optional[Callable[..., nn.Module]] = nn.BatchNorm
    activation_layer: Callable = nn.silu
    dilation: int = 1
    bias: Optional[bool] = None
    dtype: Any = jnp.float32
    conv_kernel_init: Callable = nn.initializers.kaiming_normal()

    @nn.compact
    def __call__(self, x, train: bool = True):

        if type(self.kernel_size) == int:
            kernel_size = (self.kernel_size,) * 2
        else:
            kernel_size = self.kernel_size

        if self.padding is None:
            # self.padding = (self.kernel_size - 1) // 2 * self.dilation
            padding = "SAME"
        if self.bias is None:
            bias = self.norm_layer is None

        x = nn.Conv(self.out_channels, kernel_size,
                    strides=self.stride, padding=padding,
                    kernel_dilation=self.dilation, feature_group_count=self.groups,
                    use_bias=bias, dtype=self.dtype, kernel_init=self.conv_kernel_init)(x)

        if self.norm_layer:
            if self.norm_layer == nn.BatchNorm:
                norm = partial(self.norm_layer,
                               use_running_average=not train,
                               momentum=0.9,
                               epsilon=1e-5,
                               dtype=self.dtype)
            else:
                norm = partial(self.norm_layer, dtype=self.dtype)
            x = norm()(x)
        if self.activation_layer:
            x = self.activation_layer(x)
        x = jnp.asarray(x, dtype=self.dtype)
        return x



from jax import lax, numpy as jnp


def hz2mel(f: float):
  """Hertz to mel scale conversion.

  Converts a frequency from hertz to mel values.

  Args:
    f: a frequency in Hz

  Returns:
    value of f on a mel scale
  """
  return 2595 * jnp.log10(1 + f / 700)


def mel2hz(m: float):
  """Mel to hertz conversion.

  Converts a frequency from mel to hertz values.

  Args:
    m: a frequency in mels

  Returns:
    value of m on the linear frequency scale
  """
  return 700 * (jnp.power(10, m / 2595) - 1)


def conv_dimension_numbers(input_shape):
    """
    Computes the dimension numbers based on the input shape.
    copied from flax for easy access
    """
    ndim = len(input_shape)
    lhs_spec = (0, ndim - 1) + tuple(range(1, ndim - 1))
    rhs_spec = (ndim - 1, ndim - 2) + tuple(range(0, ndim - 2))
    out_spec = lhs_spec
    return lax.ConvDimensionNumbers(lhs_spec, rhs_spec, out_spec)



"""
An implementation of the EfficientNet CNN architecture in jax

Based on the simplified implementation of EfficientNet as found in torchvision (https://pytorch.org/vision/main/_modules/torchvision/models/efficientnet.html)

Written/modified for audax by / Copyright 2022, Sarthak Yadav
"""
import copy
import jax
import math
from typing import Callable, Optional, Any, List, Sequence
from jax import numpy as jnp
from flax import linen as nn
from functools import partial
from .utils import _make_divisible
from .layers.efficientnet_helpers import SqueezeAndExcitation, ConvNormActivation
from .layers.drop import DropPath


# __all__ = [
#     "EfficientNet",
#     "efficientnet_b0",
#     "efficientnet_b1",
#     "efficientnet_b2",
#     "efficientnet_b3",
#     "efficientnet_b4",
#     "efficientnet_b5",
#     "efficientnet_b6",
#     "efficientnet_b7",
# ]


class MBConvConfig:
    # Stores information listed at Table 1 of the EfficientNet paper
    def __init__(
            self,
            expand_ratio: float,
            kernel: int,
            stride: int,
            input_channels: int,
            out_channels: int,
            num_layers: int,
            width_mult: float,
            depth_mult: float,
    ) -> None:
        self.expand_ratio = expand_ratio
        self.kernel = kernel
        self.stride = stride
        self.input_channels = self.adjust_channels(input_channels, width_mult)
        self.out_channels = self.adjust_channels(out_channels, width_mult)
        self.num_layers = self.adjust_depth(num_layers, depth_mult)

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        s += "expand_ratio={expand_ratio}"
        s += ", kernel={kernel}"
        s += ", stride={stride}"
        s += ", input_channels={input_channels}"
        s += ", out_channels={out_channels}"
        s += ", num_layers={num_layers}"
        s += ")"
        return s.format(**self.__dict__)

    @staticmethod
    def adjust_channels(channels: int, width_mult: float, min_value: Optional[int] = None) -> int:
        return _make_divisible(channels * width_mult, 8, min_value)

    @staticmethod
    def adjust_depth(num_layers: int, depth_mult: float):
        return int(math.ceil(num_layers * depth_mult))


class MBConv(nn.Module):
    cnf: MBConvConfig
    drop_path_prob: float
    norm_layer: Callable[..., nn.Module]
    se_layer: Callable[..., nn.Module] = SqueezeAndExcitation
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, inputs, train: bool = True):
        if not (1 <= self.cnf.stride <= 2):
            raise ValueError("wrong stride value")

        input_channels = inputs.shape[-1]
        use_res_connect = self.cnf.stride == 1 and input_channels == self.cnf.out_channels
        activation_layer = nn.silu

        expanded_channels = self.cnf.adjust_channels(input_channels, self.cnf.expand_ratio)
        outputs = inputs
        if expanded_channels != input_channels:
            outputs = ConvNormActivation(
                expanded_channels,
                kernel_size=1,
                norm_layer=self.norm_layer,
                activation_layer=activation_layer
            )(outputs, train=train)

        # depthwise
        outputs = ConvNormActivation(
            expanded_channels,
            kernel_size=self.cnf.kernel,
            stride=self.cnf.stride,
            groups=expanded_channels,
            norm_layer=self.norm_layer,
            activation_layer=activation_layer,
        )(outputs, train=train)

        # squeeze and excitation
        squeeze_channels = max(1, input_channels // 4)
        outputs = self.se_layer(expanded_channels, squeeze_channels, activation=activation_layer)(outputs)

        # project
        outputs = ConvNormActivation(
            self.cnf.out_channels, kernel_size=1,
            norm_layer=self.norm_layer, activation_layer=None
        )(outputs, train=train)

        if use_res_connect:
            if self.drop_path_prob != 0.:
                outputs = DropPath(self.drop_path_prob)(outputs, train=train)
            outputs = outputs + inputs
        return outputs


class EfficientNet(nn.Module):
    inverted_residual_setting: List[MBConvConfig]
    drop_rate: float
    drop_path_prob: float = 0.2
    num_classes: Optional[int] = 1000
    block: Optional[Callable[..., nn.Module]] = None
    norm_layer: Optional[Callable[..., nn.Module]] = None
    global_pool: str = "avg"
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, inputs, train: bool = True):
        if not self.inverted_residual_setting:
            raise ValueError("The inverted_residual_setting should not be empty")
        elif not (
                isinstance(self.inverted_residual_setting, Sequence)
                and all([isinstance(s, MBConvConfig) for s in self.inverted_residual_setting])
        ):
            raise TypeError("The inverted_residual_setting should be List[MBConvConfig]")

        if self.block is None:
            block = MBConv
        else:
            block = self.block

        if self.norm_layer is None:
            norm_layer = nn.BatchNorm
        else:
            norm_layer = self.norm_layer

        activation_layer = nn.silu

        firstconv_output_channels = self.inverted_residual_setting[0].input_channels
        outputs = ConvNormActivation(firstconv_output_channels, kernel_size=3, stride=2,
                                     norm_layer=norm_layer, activation_layer=activation_layer)(inputs, train=train)

        total_stage_blocks = sum(cnf.num_layers for cnf in self.inverted_residual_setting)
        stage_block_id = 0

        for cnf in self.inverted_residual_setting:
            for ix in range(cnf.num_layers):

                block_cnf = copy.copy(cnf)

                if ix > 0:
                    block_cnf.input_channels = block_cnf.out_channels
                    block_cnf.stride = 1

                sd_prob = self.drop_path_prob * float(stage_block_id) / total_stage_blocks

                outputs = block(block_cnf, norm_layer=norm_layer, drop_path_prob=sd_prob, dtype=self.dtype)(outputs,
                                                                                                            train=train)
                stage_block_id += 1

        lastconv_output_channels = 4 * self.inverted_residual_setting[-1].out_channels

        outputs = ConvNormActivation(lastconv_output_channels, kernel_size=1, norm_layer=norm_layer,
                                     activation_layer=activation_layer, dtype=self.dtype)(outputs, train=train)

        if self.global_pool == "avg":
            outputs = jnp.mean(outputs, axis=(1, 2))
        elif self.global_pool == "max":
            outputs = jnp.max(outputs, axis=(1, 2))
        else:
            raise ValueError("Unsupported pool type value : {}".format(self.global_pool))

        if self.num_classes is not None:
            if self.drop_rate != 0.:
                outputs = nn.Dropout(self.drop_rate, deterministic=not train, name="fc_dropout")(outputs)
            outputs = nn.Dense(self.num_classes, name="fc")(outputs)

        return outputs


def _efficientnet(
        arch: str,
        width_mult: float,
        depth_mult: float,
        dropout: float,
        **kwargs: Any,
) -> EfficientNet:
    bneck_conf = partial(MBConvConfig, width_mult=width_mult, depth_mult=depth_mult)
    inverted_residual_setting = [
        bneck_conf(1, 3, 1, 32, 16, 1),
        bneck_conf(6, 3, 2, 16, 24, 2),
        bneck_conf(6, 5, 2, 24, 40, 2),
        bneck_conf(6, 3, 2, 40, 80, 3),
        bneck_conf(6, 5, 1, 80, 112, 3),
        bneck_conf(6, 5, 2, 112, 192, 4),
        bneck_conf(6, 3, 1, 192, 320, 1),
    ]
    model = EfficientNet(inverted_residual_setting, dropout, **kwargs)
    return model


def efficientnet_b0(**kwargs: Any):
    return _efficientnet("efficientnet_b0", 1.0, 1.0, 0.2, **kwargs)


def efficientnet_b1(**kwargs: Any) -> EfficientNet:
    return _efficientnet("efficientnet_b1", 1.0, 1.1, 0.2, **kwargs)


def efficientnet_b2(**kwargs: Any) -> EfficientNet:
    return _efficientnet("efficientnet_b2", 1.1, 1.2, 0.3, **kwargs)


def efficientnet_b3(**kwargs: Any) -> EfficientNet:
    return _efficientnet("efficientnet_b3", 1.2, 1.4, 0.3, **kwargs)


def efficientnet_b4(**kwargs: Any) -> EfficientNet:
    return _efficientnet("efficientnet_b4", 1.4, 1.8, 0.4, **kwargs)


def efficientnet_b5(**kwargs: Any) -> EfficientNet:
    return _efficientnet(
        "efficientnet_b5",
        1.6,
        2.2,
        0.4,
        norm_layer=partial(nn.BatchNorm, epsilon=0.001, momentum=0.01),
        **kwargs,
    )


def efficientnet_b6(**kwargs: Any) -> EfficientNet:
    return _efficientnet(
        "efficientnet_b6",
        1.8,
        2.6,
        0.5,
        norm_layer=partial(nn.BatchNorm, epsilon=0.001, momentum=0.01),
        **kwargs,
    )


def efficientnet_b7(**kwargs: Any) -> EfficientNet:
    return _efficientnet(
        "efficientnet_b7",
        2.0,
        3.1,
        0.5,
        norm_layer=partial(nn.BatchNorm, epsilon=0.001, momentum=0.01),
        **kwargs,
    )



import enum


@enum.unique
class TrainingMode(enum.Enum):
    """Look up for similarity measure in contrastive model."""
    MULTICLASS = "multiclass"
    MULTILABEL = "multilabel"
    CONTRASTIVE = "contrastive"
    COLA = "cola"


@enum.unique
class DataSplit(enum.Enum):
    """Look up for similarity measure in contrastive model."""
    TRAIN = "train"
    EVAL = "eval"
    VAL = "val"


@enum.unique
class Features(enum.Enum):
    """Look up for similarity measure in contrastive model."""
    RAW = "raw"
    SPECTROGRAM = "spectrogram"
    LOGMEL = "log_mel"



"""
functional utils for audio feature extraction

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import jax
import warnings
import math
from jax import numpy as jnp, lax
from typing import Optional, Callable, Any
from .stft import stft
from .helpers import batch_pad


def _hz_to_mel(freq: float, mel_scale: str = "htk") -> float:
    r"""Convert Hz to Mels.

    Args:
        freqs (float): Frequencies in Hz
        mel_scale (str, optional): Scale to use: ``htk`` or ``slaney``. (Default: ``htk``)

    Returns:
        mels (float): Frequency in Mels
    """
    if mel_scale not in ['slaney', 'htk']:
        raise ValueError('mel_scale should be one of "htk" or "slaney".')

    if mel_scale == "htk":
        return 2595.0 * math.log10(1.0 + (freq / 700.0))

    # Fill in the linear part
    f_min = 0.0
    f_sp = 200.0 / 3

    mels = (freq - f_min) / f_sp
    min_log_hz = 1000.0
    min_log_mel = (min_log_hz - f_min) / f_sp
    logstep = math.log(6.4) / 27.0

    if freq >= min_log_hz:
        mels = min_log_mel + math.log(freq / min_log_hz) / logstep
    return mels


def _mel_to_hz(mels: jnp.array, mel_scale: str = "htk") -> jnp.array:
    """Convert mel bin numbers to frequencies.

    Args:
        mels (jnp.array): Mel frequencies
        mel_scale (str, optional): Scale to use: ``htk`` or ``slaney``. (Default: ``htk``)

    Returns:
        freqs (Tensor): Mels converted in Hz
    """
    if mel_scale not in ['slaney', 'htk']:
        raise ValueError('mel_scale should be one of "htk" or "slaney".')
    if mel_scale == "htk":
        return 700.0 * (10.0 ** (mels / 2595.0) - 1.0)
    # Fill in the linear scale
    f_min = 0.0
    f_sp = 200.0 / 3
    freqs = f_min + f_sp * mels

    # And now the nonlinear scale
    min_log_hz = 1000.0
    min_log_mel = (min_log_hz - f_min) / f_sp
    logstep = math.log(6.4) / 27.0

    log_t = (mels >= min_log_mel)
    freqs.at[log_t].set(min_log_hz * jnp.exp(logstep * (mels[log_t] - min_log_mel)))
    return freqs


def _create_triangular_filterbank(
        all_freqs: jnp.array,
        f_pts: jnp.array,
) -> jnp.array:
    """Create a triangular filter bank.

    Args:
        all_freqs (Tensor): STFT freq points of size (`n_freqs`).
        f_pts (Tensor): Filter mid points of size (`n_filter`).

    Returns:
        fb (Tensor): The filter bank of size (`n_freqs`, `n_filter`).
    """
    # Adopted from Librosa
    # calculate the difference between each filter mid point and each stft freq point in hertz
    f_diff = f_pts[1:] - f_pts[:-1]  # (n_filter + 1)
    slopes = f_pts[jnp.newaxis, Ellipsis] - all_freqs[Ellipsis, jnp.newaxis]  # (n_freqs, n_filter + 2)
    # create overlapping triangles
    down_slopes = (-1.0 * slopes[:, :-2]) / f_diff[:-1]  # (n_freqs, n_filter)
    up_slopes = slopes[:, 2:] / f_diff[1:]  # (n_freqs, n_filter)
    fb = jnp.maximum(0., jnp.minimum(down_slopes, up_slopes))
    return fb


def melscale_fbanks(
        n_freqs: int,
        n_mels: int,
        sample_rate: int,
        f_min: float,
        f_max: Optional[float] = None,
        norm: Optional[str] = None,
        mel_scale: str = "htk",
) -> jnp.array:
    r"""Create a frequency bin conversion matrix.

    Args:
        n_freqs (int): Number of frequencies to highlight/apply
        f_min (float): Minimum frequency (Hz)
        f_max (float): Maximum frequency (Hz)
        n_mels (int): Number of mel filterbanks
        sample_rate (int): Sample rate of the audio waveform
        norm (str or None, optional): If 'slaney', divide the triangular mel weights by the width of the mel band
            (area normalization). (Default: ``None``)
        mel_scale (str, optional): Scale to use: ``htk`` or ``slaney``. (Default: ``htk``)

    Returns:
        Tensor: Triangular filter banks (fb matrix) of size (``n_freqs``, ``n_mels``)
        meaning number of frequencies to highlight/apply to x the number of filterbanks.
        Each column is a filterbank so that assuming there is a matrix A of
        size (..., ``n_freqs``), the applied result would be
        ``A * melscale_fbanks(A.size(-1), ...)``.

    """
    f_max = f_max if f_max is not None else float(sample_rate // 2)
    assert f_min <= f_max, 'Require f_min: {} < f_max: {}'.format(f_min, f_max)
    if norm is not None and norm != "slaney":
        raise ValueError("norm must be one of None or 'slaney'")
    all_freqs = jnp.linspace(0, sample_rate // 2, n_freqs)

    # calculate mel freq bins
    m_min = _hz_to_mel(f_min, mel_scale=mel_scale)
    m_max = _hz_to_mel(f_max, mel_scale=mel_scale)

    m_pts = jnp.linspace(m_min, m_max, n_mels + 2)
    f_pts = _mel_to_hz(m_pts, mel_scale=mel_scale)

    # create filterbank
    fb = _create_triangular_filterbank(all_freqs, f_pts)

    if norm is not None and norm == "slaney":
        # Slaney-style mel is scaled to be approx constant energy per channel
        enorm = 2.0 / (f_pts[2:n_mels + 2] - f_pts[:n_mels])
        fb *= enorm[jnp.newaxis, Ellipsis]

    # if (fb.max(axis=0) == 0.).any():
    #     warnings.warn(
    #         "At least one mel filterbank has all zero values. "
    #         f"The value for `n_mels` ({n_mels}) may be set too high. "
    #         f"Or, the value for `n_freqs` ({n_freqs}) may be set too low."
    #     )

    return fb


def spectrogram(
        inputs: jnp.array,
        pad: int,
        window: jnp.array,
        n_fft: int,
        hop_length: int,
        win_length: int,
        power: Optional[float] = 2.0,
        normalized: bool = False,
        center: bool = True,
        onesided: bool = True,
        return_complex: bool = True
) -> jnp.array:
    r"""Create a spectrogram or a batch of spectrograms from a raw audio signal.
        The spectrogram can be either magnitude-only or complex.

        Args:
            inputs (Tensor): Tensor of audio of dimension `(..., time)`
            pad (int): Two sided padding of signal
            window (Tensor): Window tensor that is applied/multiplied to each frame/window
            n_fft (int): Size of FFT
            hop_length (int): Length of hop between STFT windows
            win_length (int): Window size
            power (float or None): Exponent for the magnitude spectrogram,
                (must be > 0) e.g., 1 for energy, 2 for power, etc.
                If None, then the complex spectrum is returned instead.
            normalized (bool): Whether to normalize by magnitude after stft
            center (bool, optional): whether to pad :attr:`waveform` on both sides so
                that the :math:`t`-th frame is centered at time :math:`t \times \text{hop\_length}`.
                Default: ``True``
            pad_mode (string, optional): controls the padding method used when
                :attr:`center` is ``True``. Default: ``"reflect"``
            onesided (bool, optional): controls whether to return half of results to
                avoid redundancy. Default: ``True``
            return_complex (bool, optional):
                Indicates whether the underlying complex valued spectrogram is to be returned.
                If False, absolute value of the spectrogram is returned
                This argument is only effective when ``power=None``. It is ignored for
                cases where ``power`` is a number as in those cases, the returned tensor is
                power spectrogram, which is a real-valued tensor.

        Returns:
            Tensor: Dimension `(..., freq, time)`, freq is
            ``n_fft // 2 + 1`` and ``n_fft`` is the number of
            Fourier bins, and time is the number of window hops (n_frame).
        """
    if power is None and not return_complex:
        raise ValueError(f"in correct combination of power(={power}) and return_complex(={return_complex}) provided.")

    if pad > 0:
        inputs = batch_pad(inputs, pad, "zeros")

    spec_f = stft(
        inputs,
        n_fft,
        hop_length,
        win_length,
        window,
        center=center,
        onesided=onesided
    )

    if normalized:
        spec_f /= jnp.sqrt(jnp.sum(jnp.power(window, 2.)))

    if power is not None:
        if power == 1.0:
            return jnp.abs(spec_f)
        return jnp.power(jnp.abs(spec_f), power)
    if not return_complex:
        return jnp.abs(spec_f)
    return spec_f


def apply_melscale(spectrogram: jnp.array,
                   melscale_filterbank: jnp.array,
                   precision: Any = lax.Precision.HIGHEST) -> jnp.array:
    r"""

    Args:
        spectrogram (jnp.array): A spectrogram STFT of dimension (..., time, freq)
        melscale_filterbank (jnp.array): MelScale filterbank of dimension (``n_freqs``, ``n_mels``)
        precision (jax.lax.Precision): Precision enum. Default value is  Precision.HIGHEST, which corresponds to FP32

    Returns
        jnp.array: A melspectrogram of dimension (..., time, ``n_mels``)
    """
    melspec = jnp.matmul(spectrogram, melscale_filterbank, precision=precision)
    return melspec



import enum
from typing import Optional
import jax


def _make_divisible(v: float, divisor: int, min_value: Optional[int] = None) -> int:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


@enum.unique
class SimilarityMeasure(enum.Enum):
    """Look up for similarity measure in contrastive model."""
    DOT = "dot_product"
    BILINEAR = "bilinear_product"


def l2_normalize(x, axis=None, eps=1e-12):
    """
    l2 normalize the input on axis dimension
    """
    return x * jax.lax.rsqrt((x * x).sum(axis=axis, keepdims=True) + eps)



"""
General classifier wrapper for audax models

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import jax.numpy as jnp
from jax import lax, random
import flax.linen as nn
from typing import Callable, Any, Optional


Dtype = Any


class Classifier(nn.Module):
    model_cls: Callable
    num_classes: int
    frontend_cls: Optional[Callable] = None
    drop_rate: float = 0.0
    spec_aug: Optional[Callable] = None
    dtype: Dtype = jnp.float32

    @nn.compact
    def __call__(self, inputs, train: bool = True):
        outputs = inputs
        if self.frontend_cls is not None:
            outputs = self.frontend_cls(dtype=jnp.float32, name="frontend")(outputs)
            outputs = outputs[Ellipsis, jnp.newaxis]
            outputs = outputs.astype(self.dtype)

            if self.spec_aug is not None and train:
                print("Doing spec_aug")
                spec_aug_rng = self.make_rng('spec_aug')
                outputs, _ = self.spec_aug(outputs, spec_aug_rng)

        outputs = self.model_cls(num_classes=None, dtype=self.dtype, name="encoder")(outputs, train=train)
        if self.drop_rate != 0:
            outputs = nn.Dropout(self.drop_rate, deterministic=not train, name="fc_dropout")(outputs)
        outputs = nn.Dense(self.num_classes, name="fc")(outputs)
        return outputs



"""
Helper functions for evaluating a given supervised "Classifier" model

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import json
import functools
import os
import time
from typing import Any
import tqdm
from absl import logging
from clu import metric_writers
from clu import periodic_actions
import flax
from flax import jax_utils
from flax import optim
from flax.training import checkpoints
from flax.training import common_utils
from flax.training import train_state
import jax
from jax import lax
import jax.numpy as jnp
from jax import random
import ml_collections
from numpy import var
import numpy as np
import optax
import tensorflow as tf
from sklearn.metrics import average_precision_score, accuracy_score
from ..transforms import mixup
from . import metrics_helper
from .trainstate import TrainState_v2
from jax.config import config


try:
    # Disable all GPUS
    tf.config.set_visible_devices([], 'GPU')
    visible_devices = tf.config.get_visible_devices()
    for device in visible_devices:
        assert device.device_type != 'GPU'
except:
    # Invalid device or cannot modify virtual devices once initialized.
    pass

from .misc import TrainingMode, Features, DataSplit
from . import training_utilities
from .data_v2.helpers import prepare_datasets_v2
from .. import models
from audax import frontends
from .train_supervised import create_model


def forward(state, batch):
    variables = {
        'params': state.get_all_params,                    # absolutely ok to just use state.get_all_params here
        'batch_stats': state.batch_stats
    }
    logits = state.apply_fn(
        variables, batch['audio'], train=False, mutable=False)
    return logits


def load_pretrained(rng, config, model, learning_rate_fn, workdir, prefix):
    dynamic_scale = None
    platform = jax.local_devices()[0].platform
    if config.half_precision and platform == 'gpu':
        dynamic_scale = optim.DynamicScale()
    else:
        dynamic_scale = None
    params, batch_stats, rng_keys = training_utilities.initialize(rng, config.input_shape, model)

    pretrained_state_dict = checkpoints.restore_checkpoint(workdir, None, prefix=prefix)

    pretrained_params = pretrained_state_dict['params']
    pretrained_frozen_params = pretrained_state_dict['frozen_params']
    pretrained_batch_stats = pretrained_state_dict['batch_stats']

    tx = optax.adamw(
        learning_rate=learning_rate_fn,
        weight_decay=config.opt.weight_decay
    )
    state = TrainState_v2.create(
        apply_fn=model.apply,
        params=pretrained_params,
        frozen_params=pretrained_frozen_params,
        tx=tx,
        batch_stats=pretrained_batch_stats,
        aux_rng_keys=rng_keys,
        dynamic_scale=dynamic_scale)

    return state


def evaluate(workdir: str,
             eval_signal_duration="AUTO",
             eval_manifest_override=None,
             eval_steps_override=None):
    config = training_utilities.read_config_from_json(workdir)
    if config.batch_size % jax.device_count() > 0:
        raise ValueError('Batch size must be divisible by the number of devices')
    # doing one sample at a time. Better supports VoxCeleb and other variable length sequences
    local_batch_size = 1 // jax.process_count()
    logging.info("Process count: {}".format(jax.process_count()))
    # eval is done on single device
    device = config.get("device", 1)
    if device:
        devices = [jax.local_devices()[device]]
    else:
        devices = jax.local_devices()
    platform = devices[0].platform
    if config.half_precision:
        if platform == 'tpu':
            input_dtype = tf.bfloat16
        else:
            input_dtype = tf.float16
    else:
        input_dtype = tf.float32
    mode = TrainingMode(config.model.type)
    if eval_signal_duration == "AUTO":
        if config.data.dataset_name == "audioset":
            config.audio_config.min_duration = 10.
        elif config.data.dataset_name == "speechcommandsv2":
            config.audio_config.min_duration = 1.
        elif config.data.dataset_name == "voxceleb1":
            config.audio_config.min_duration = 10.
        else:
            raise ValueError(f"Unsupported dataset {config.data.dataset_name} for eval_signal_duration == 'AUTO'")
    elif type(eval_signal_duration) == float and eval_signal_duration >= 1.0:
        config.audio_config.min_duration = eval_signal_duration
    else:
        raise ValueError(f"Unsupported dataset {config.data.dataset_name} for eval_signal_duration == 'AUTO'")
    if eval_manifest_override is not None:
        assert os.path.exists(eval_manifest_override), f"{eval_manifest_override} doesn't exist"
        logging.info("Overriding eval_manifest path {} in config file with {}".format(
            config.data.eval_manifest, eval_manifest_override
        ))
        if eval_steps_override is None or eval_steps_override == 0:
            raise ValueError(f"Incorrect value for eval_steps_override: {eval_steps_override}")
        config.data.eval_manifest = eval_manifest_override
        config.data.eval_samples = eval_steps_override

    rng = random.PRNGKey(0)
    _, eval_iter = prepare_datasets_v2(config, local_batch_size, input_dtype=input_dtype)
    eval_iter = training_utilities.create_input_iter(eval_iter, devices=devices)

    if config.data.jax_transforms:
        tfs = training_utilities.get_feature_functions(config)
        if len(tfs) != 0:
            p_feature_extract_fn = jax.pmap(
                functools.partial(
                    training_utilities.apply_audio_transforms, transforms=tfs,
                    dtype=training_utilities.get_dtype(config.half_precision),
                    normalize=config.data.get("normalize_batch", False)
                ), axis_name='batch', devices=devices)
        else:
            p_feature_extract_fn = None
    else:
        p_feature_extract_fn = None
    model_cls, frontend_cls = training_utilities.get_model_frontend_cls(config)
    model = create_model(
        model_cls=model_cls, half_precision=config.half_precision,
        frontend_cls=frontend_cls,
        num_classes=config.model.num_classes,
        spec_aug=None,
        drop_rate=config.model.get("fc_drop_rate", 0.))

    # placeholder to just load the thing
    learning_rate_fn = training_utilities.create_learning_rate_fn(
        config, 0.1, 100)
    # state = training_utilities.create_train_state(rng, config, model, learning_rate_fn)
    # state = checkpoints.restore_checkpoint(workdir, state, prefix="best_")
    state = load_pretrained(rng, config, model, learning_rate_fn, workdir, prefix="best_")
    state = jax_utils.replicate(state, devices=devices)
    p_forward = jax.pmap(functools.partial(forward),
                           axis_name='batch', devices=devices)
    if config.steps_per_eval == -1:
        num_validation_examples = config.data.eval_samples
        steps_per_eval = num_validation_examples // 1
    else:
        steps_per_eval = config.steps_per_eval
    eval_logits = []
    eval_labels = []
    for _ in tqdm.tqdm(range(steps_per_eval)):
        eval_batch = next(eval_iter)
        if p_feature_extract_fn:
            eval_batch['audio'] = p_feature_extract_fn(eval_batch['audio'])
        # print(eval_batch['audio'].shape)
        logits = p_forward(state, eval_batch)
        # print(logits.shape)
        eval_logits.append(logits)
        eval_labels.append(eval_batch['label'])
    logging.info("Concatenating predictions and labels..")
    eval_logits = jnp.concatenate([jax.device_get(x) for x in eval_logits])
    eval_labels = jnp.concatenate([jax.device_get(x) for x in eval_labels])
    eval_logits = eval_logits.reshape(-1, eval_logits.shape[-1]).astype('float32')
    eval_labels = eval_labels.reshape(-1, eval_labels.shape[-1]).astype("float32")
    fp = open(os.path.join(workdir, "results.txt"), "w")
    if mode == TrainingMode.MULTILABEL:
        # macro_mAP = metrics_helper.calculate_mAP(eval_logits, eval_labels)
        stats = metrics_helper.calculate_stats(eval_logits, eval_labels)
        mAP = np.mean([stat['AP'] for stat in stats])
        mAUC = np.mean([stat['auc'] for stat in stats])
        dp = metrics_helper.d_prime(mAUC)
        s = "mAP: {:.5f}\n".format(mAP)
        s += "mAUC: {:.5f}\n".format(mAUC)

        s += "dprime: {:.5f}".format(dp)
    elif mode == TrainingMode.MULTICLASS:
        acc = accuracy_score(y_true=np.argmax(np.asarray(eval_labels), axis=1), 
                             y_pred=np.argmax(np.asarray(eval_logits), axis=1))
        s = "Accuracy: {:.4f}".format(acc)
    print(s)
    fp.write(s)
    fp.close()



"""
some helpers utility functions

_conv_dimension_numbers is taken from flax (https://flax.readthedocs.io/en/latest/_modules/flax/linen/linear.html)

Written/hacked together for audax by / Copyright 2022, Sarthak Yadav
"""
import jax
from functools import partial
from jax import numpy as jnp, lax


def batch_pad(xs, pad: int, mode="reflect"):
    if xs.ndim == 2:
        pad_arg = [(0, 0), (pad, pad)]
    elif xs.ndim == 3:
        pad_arg = [(0, 0), (pad, pad), (0, 0)]
    else:
        pad_arg = [(pad, pad)]
    xs = jnp.pad(xs, pad_arg, mode=mode)
    return xs


def _conv_dimension_numbers(input_shape):
    """Computes the dimension numbers based on the input shape."""
    ndim = len(input_shape)
    lhs_spec = (0, ndim - 1) + tuple(range(1, ndim - 1))
    rhs_spec = (ndim - 1, ndim - 2) + tuple(range(0, ndim - 2))
    out_spec = lhs_spec
    return lax.ConvDimensionNumbers(lhs_spec, rhs_spec, out_spec)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="audax",
    version="0.0.3",
    author='Sarthak Yadav',
    description="audio ML for Jax",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SarthakYadav/audio_data_utils",
    # package_dir={"": ""},
    packages=[
        "audax",
        "audax.commons",
        "audax.core",
        "audax.frontends",
        "audax.models",
        "audax.models.layers",
        "audax.transforms",
        "audax.training_utils",
        "audax.training_utils.data_v2",
    ],
    python_requires=">=3.8"
)



"""
Implementation of DropPath (Stochastic Depth) regularization

Inspired by the PyTorch implementation in timm (https://github.com/rwightman/pytorch-image-models)
by Ross Wightman, 2022

Modifications and additions for audax by / Copyright 2022, Sarthak Yadav
"""

from flax import linen as nn
from jax import lax, numpy as jnp, random
from typing import Any


def drop_path(x: jnp.array, rng, drop_prob: float = 0.) -> jnp.array:
    if drop_prob == 0.:
        return x
    keep_prob = 1 - drop_prob
    mask = random.bernoulli(key=rng, p=keep_prob, shape=(x.shape[0],) + (1,)*(x.ndim-1))
    mask = jnp.broadcast_to(mask, x.shape)
    return lax.select(mask, x / keep_prob, jnp.zeros_like(x))


class DropPath(nn.Module):
    rate: float = 0.

    @nn.compact
    def __call__(self, x, train: bool = False, rng: Any = None):
        if not train or self.rate == 0.:
            return x
        if rng is None:
            rng = self.make_rng('drop_path')
        return drop_path(x, rng, drop_prob=self.rate)



"""
Implementation for the paper

Park, D.S., Chan, W., Zhang, Y., Chiu, C.C., Zoph, B., Cubuk, E.D. and Le, Q.V., 2019.
"Specaugment: A simple data augmentation method for automatic speech recognition."
arXiv preprint arXiv:1904.08779.

Written for audax by / Copyright 2022, Sarthak Yadav
"""
from typing import Any
import jax.numpy as jnp
from jax import random, jit
import flax.linen as nn


class SpecAugment:
    """
    SpecAugment, without time-stretching
    """
    def __init__(self, freq_param: int, time_param: int, num_masks: int = 2):
        self.freq_param = freq_param
        self.time_param = time_param
        self.num_masks = num_masks

    def augment(self, inputs, rng):
        # if rng is None:
        #     rng = self.make_rng('spec_augment')
        rng, *keys = random.split(rng, 3)

        def mask_along_axis_iid(key_index, inputs: jnp.array, mask_param, axis, mask_value=0.):
            """
            :param inputs: Input tensor of shape (batch, time, freq, 1)
            :param mask_param:
            :param mask_value:
            :param axis:
            :return:
            """
            # if axis not in [1, 2]:
            #     raise ValueError('Only Frequency and Time masking are supported')

            dtype = inputs.dtype

            shape_ = (inputs.shape[0], inputs.shape[-1])

            value = random.uniform(keys[key_index], shape_) * mask_param
            min_value = random.uniform(keys[key_index], shape_) * (inputs.shape[axis] - value)
            mask_start = min_value.reshape(-1, 1, 1, 1)
            mask_end = (min_value + value).reshape(-1, 1, 1, 1)
            mask = jnp.arange(0, inputs.shape[axis])
            if axis == 1:
                inputs = inputs.transpose(0, 2, 3, 1)
            else:
                inputs = inputs.transpose(0, 1, 3, 2)
            o = (mask >= mask_start) & (mask < mask_end)
            o = jnp.repeat(o, inputs.shape[1], axis=1)
            inputs = jnp.where(o, 0, inputs)# inputs.at[o].set(mask_value)

            if axis == 1:
                inputs = inputs.transpose(0, 3, 1, 2)
            else:
                inputs = inputs.transpose(0, 1, 3, 2)
            return inputs

        outputs = inputs
        if self.freq_param != 0:
            outputs = mask_along_axis_iid(0, outputs, self.freq_param, 2)
        if self.time_param != 0:
            outputs = mask_along_axis_iid(1, outputs, self.time_param, 1)
        return outputs, rng

    def __call__(self, inputs, rng):
        outputs = inputs
        for _ in range(self.num_masks):
            outputs, rng = self.augment(outputs, rng)
        return outputs, rng



"""
Helper utilities for parsing experiment configs and creating data splits.

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import ml_collections
import pandas as pd
import functools
from . import dataset
from . import transforms
from ..misc import DataSplit, TrainingMode, Features


def get_tfrecord_parser(config: ml_collections.ConfigDict, cropper,
                        label_parser_func, is_val=False,
                        is_contrastive=False):
    desired_seq_len = int(config.audio_config.sample_rate * config.audio_config.min_duration)
    if is_contrastive and config.data.reader != "tfio":
        raise NotImplementedError("Only tfio data reader supports contrastive data parsing." +
                                  "Provided config.data.reader == {}".format(config.data.reader))
    if config.data.reader == "tfio":
        print("using tfio bruh")
        if is_contrastive:
            parser_fn = functools.partial(dataset.parse_tf_record_tfio_contrastive,
                                          label_parser=label_parser_func,
                                          cropper=cropper,
                                          noise=config.model.get("contrastive_noise", 0.001),
                                          normalize_waveform=config.audio_config.get("normalize_waveform", False),
                                          seg_length=desired_seq_len * 2)
        else:
            parser_fn = functools.partial(dataset.parse_tfrecord_fn_tfio,
                                          label_parser=label_parser_func,
                                          cropper=cropper,
                                          normalize_waveform=config.audio_config.get("normalize_waveform", False),
                                          seg_length=desired_seq_len)
    else:
        parser_fn = functools.partial(dataset.parse_tfrecord_fn_v2,
                                      label_parser=label_parser_func,
                                      cropper=cropper,
                                      seg_length=desired_seq_len)
    return parser_fn


def prepare_datasets_v2(config: ml_collections.ConfigDict, batch_size, input_dtype):
    train_files = pd.read_csv(config.data.tr_manifest)['files'].values
    val_files = pd.read_csv(config.data.eval_manifest)['files'].values

    desired_seq_len = int(config.audio_config.sample_rate * config.audio_config.min_duration)

    label_parser_func = functools.partial(transforms.label_parser, mode=config.model.type,
                                          num_classes=config.model.num_classes)

    random_cropper = functools.partial(transforms.random_crop_signal, slice_length=desired_seq_len)
    center_cropper = functools.partial(transforms.center_crop_signal, slice_length=desired_seq_len)

    if config.model.type.lower() in ["contrastive", "cola"]:
        is_contrastive = True
    else:
        is_contrastive = False
    print("prepare_datasets_v2:", config.model.type.lower(), is_contrastive)

    parse_record_train = get_tfrecord_parser(config, cropper=random_cropper,
                                             label_parser_func=label_parser_func,
                                             is_contrastive=is_contrastive)
    parse_record_val = get_tfrecord_parser(config, cropper=center_cropper,
                                           label_parser_func=label_parser_func,
                                           is_val=True,
                                           is_contrastive=is_contrastive)
    fe = None
    # TODO: remove torch transforms
    if not config.data.jax_transforms:
        from . import torch_transforms
        if config.audio_config.features == "log_mel":
            fe = torch_transforms.get_log_mel_extractor(
                sample_rate=config.audio_config.sample_rate,
                frame_length=config.audio_config.win_len,
                frame_step=config.audio_config.hop_len,
                fft_length=config.audio_config.n_fft,
                n_mels=config.audio_config.n_mels,
                fmin=config.audio_config.fmin,
                fmax=config.audio_config.fmax
            )
        elif config.audio_config.features == "spectrogram":
            fe = torch_transforms.get_spectrogram_extractor(
                frame_length=config.audio_config.win_len,
                frame_step=config.audio_config.hop_len,
                fft_length=config.audio_config.n_fft
            )
        else:
            fe = None
    if fe and not config.data.jax_transforms:
        batched_feature_extractor = functools.partial(transforms.map_torch_batched_feature_extractor,
                                                      feature_extractor=fe)
    else:
        batched_feature_extractor = None

    train_dataset = dataset.get_dataset_v2(
        train_files, batch_size, parse_example=parse_record_train, num_classes=config.model.num_classes,
        data_split=DataSplit.TRAIN, compression=config.data.get("compression", "ZLIB"),
        feature_extraction_func=batched_feature_extractor, cacheable=config.data.cacheable,
        is_contrastive=is_contrastive
    )

    val_dataset = dataset.get_dataset_v2(
        val_files, batch_size, parse_example=parse_record_val, num_classes=config.model.num_classes,
        data_split=DataSplit.EVAL, compression=config.data.get("compression", "ZLIB"),
        feature_extraction_func=batched_feature_extractor, cacheable=config.data.cacheable,
        is_contrastive=is_contrastive
    )

    if not config.data.jax_transforms:
        dtype_map_func = functools.partial(transforms.map_dtype, desired=input_dtype)
        train_dataset = train_dataset.map(dtype_map_func)
        val_dataset = val_dataset.map(dtype_map_func)
    return train_dataset, val_dataset



"""
Short time fourier transform implementation in jax

Written for audax by / Copyright 2022, Sarthak Yadav
"""
import jax
from jax import numpy as jnp, lax
from typing import Optional, Callable
from .helpers import batch_pad, _conv_dimension_numbers


def stft(inputs: jnp.array, n_fft: int, hop_length: Optional[int] = None,
         win_length: Optional[int] = None, window: Optional[jnp.array] = None,
         center: bool = True,
         onesided: Optional[bool] = None) -> jnp.array:
    r"""Short-time fourier transform
    This implementation is designed to be consistent with stft implementation from torchaudio
    and does not have all the bells and whistles from scipy.signal.

    Get's the job done though, run's on GPUs and TPUs, and produces identical output to the torchaudio implementation
    in very few lines of code.

    It uses `lax.conv_general_dilated_patches` to generate overlapping patches.

    Returns a jnp.array of size :math:`(* \times T \times N)`, where :math:`*` is the optional batch size of
    :attr:`input`, :math:`N` is the number of frequencies where STFT is applied
    and :math:`T` is the total number of frames used.

    Args:
        input (jnp.array): input signal tensor of shape (T,), (N, T) or (N, T, 1)
        n_fft (int): size of Fourier transform
        hop_length (int, optional): the distance between neighboring sliding window
            frames. Default: ``None`` (treated as equal to ``floor(n_fft / 4)``)
        win_length (int, optional): the size of window frame and STFT filter.
            Default: ``None``  (treated as equal to :attr:`n_fft`)
        window (Tensor, optional): the optional window function.
            Default: ``None`` (treated as window of all :math:`1` s)
        center (bool, optional): whether to pad :attr:`input` on both sides via reflection padding
            so that the :math:`t`-th frame is centered at time :math:`t \times \text{hop\_length}`.
            Default: ``True``
        onesided (bool, optional): controls whether to return half of results to
            avoid redundancy for real inputs.
            Default: ``True`` for real :attr:`input` and :attr:`window`, ``False`` otherwise.
    Returns:
        jnp.array, as described above
    """

    is_real = inputs.dtype == jnp.float32

    assert window.shape[-1] == win_length
    win_length = win_length if win_length is not None else n_fft
    hop_length = hop_length if hop_length is not None else n_fft // 4

    input_dims = inputs.ndim
    single_input = False
    if input_dims == 1:
        inputs = inputs.reshape(1, -1, 1)
        single_input = True
    if input_dims == 2:
        # assuming first dimension is batch, second is time
        inputs = inputs[Ellipsis, jnp.newaxis]

    if center:
        pad = int(n_fft // 2)
        inputs = batch_pad(inputs, pad=pad)

    dimension_numbers = _conv_dimension_numbers(inputs.shape)
    strided = lax.conv_general_dilated_patches(lhs=inputs, filter_shape=(win_length,),
                        window_strides=(hop_length,), padding='VALID',
                        dimension_numbers=dimension_numbers)

    # apply window
    strided = strided * window.reshape(1, 1, -1)
    output = jnp.fft.fft(strided, n=n_fft)

    if onesided is None:
        if is_real:
            onesided = True

    if onesided:
        output = output[:, :, :(n_fft // 2) + 1]

    return output
